

#include "nanojit.h"
#include "NativeInterpreterMacros.h"
#include "NativeInterpreterTypes.h"

#include <limits.h>


namespace nanojit
{

#if defined(NANOJIT_VERBOSE)
const char* getRegName(Register r) {
	return regNames[REGNUM(r)];
}
#endif

const char* getCmpOpName(nanojit::LOpcode cmpOp) {

	switch (cmpOp) {
		case LIR_eqi:
		case LIR_eqq:
		case LIR_eqf:
		case LIR_eqd:
			return "eq";

		case LIR_nei:
		case LIR_neq:
		case LIR_nef:
		case LIR_ned:
			return "ne";

		case LIR_lti:
		case LIR_ltq:
		case LIR_ltf:
		case LIR_ltd:
			return "lt";


		case LIR_gei:
		case LIR_geq:
		case LIR_gef:
		case LIR_ged:
			return "ge";

		case LIR_gti:
		case LIR_gtq:
		case LIR_gtf:
		case LIR_gtd:
			return "gt";

		case LIR_lei:
		case LIR_leq:
		case LIR_lef:
		case LIR_led:
			return "le";

		case LIR_ltui:
		case LIR_ltuq:
			return "ltu";

		case LIR_geui:
		case LIR_geuq:
			return "geu";

		case LIR_gtui:
		case LIR_gtuq:
			return "gtu";

		case LIR_leui:
		case LIR_leuq:
			return "leu";

		case LIR_unf:
		case LIR_und:
			return "un";

		default:
			NanoAssert(0); return "invalid compare op";
	}

}

const char *regNames[] = {
	"r0",  "r1",  "r2",  "r3",  "r4",  "r5",  "r6",  "r7",
	"r8",  "r9",  "r10", "r11", "r12", "r13", "r14", "r15",
	"r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
	"r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31"
};


void RegAlloc::nInitHint() {
}

RegisterMask RegAlloc::nRegCopyCandidates(Register r, RegisterMask allow) {
	NanoUnusedVariable(r);
	return allow;
}

RegisterMask RegAlloc::nInitManagedRegisters(const Config* config) {
	NanoUnusedVariable(config);
	// all registers are available
	return 0xFFFFFFFF & ~(rmask(FP));
}

RegisterMask RegAlloc::nHint(LIns* ins) {
	NanoUnusedVariable(ins);
	return 0;
}

bool RegAlloc::nCanRemat(LIns* ins) {
	return ins->isImmAny() && ins->canOptimize();
}


//////////////

void Assembler::nCreate() {
}

void Assembler::nDestroy() {
}

void Assembler::nBeginAssembly(Allocator* compilationAllocator) {
	NanoUnusedVariable(compilationAllocator);
}

void Assembler::nEndAssembly(Allocator* compilationAllocator) {
	NanoUnusedVariable(compilationAllocator);
}

void Assembler::nPatchRelativeBranch(NIns *patch, const NIns *target) {

	NIns* stream = patch;
	uint32_t op;
	READ_OPCODE(op, stream);

	if ((op == LIR_branch) || (op == LIR_branch_cond)){
		ptrdiff_t jumpOffset = target - (stream + sizeof(ptrdiff_t));
		*((ptrdiff_t*)stream) = jumpOffset;
	} else {
		NanoAssert(0); // invalid opcode for branch patching
	}
}

void Assembler::nPatchRelativeOffset(const NIns *originalAddress, NIns *opcode) {
	NanoUnusedVariable(originalAddress);
	NanoUnusedVariable(opcode);
	NanoAssert(0);
}

NIns* Assembler::nEntryPoint(RegisterMask saveMask) {
	NanoAssert(saveMask == 0);
	NanoUnusedVariable(saveMask);
	return _nIns;
}

NIns* Assembler::nExitPoint(RegisterMask restoreMask) {
	NanoAssert(restoreMask == 0);
	NanoUnusedVariable(restoreMask);

	underrunProtect(4);
	WRITE_OPCODE(LIR_exit, _nIns, "", "");
	return _nIns;
}

void Assembler::nPrologue() {
}

void Assembler::nInsertRandomNop() {
}

void Assembler::nRestore(LIns *ins, Register r) {

	if (RegAlloc::nCanRemat(ins)) {
		asm_restore_remat(ins, r);
		return;
	}

	int d = findMemFor(ins);
	LTy retType = ins->retType();


	underrunProtect(12);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);

	unsigned int op = 0;

	switch (retType) {
		case LTy_I:		op = LIR_restorei; break;
#if defined(NANOJIT_64BIT)
		case LTy_Q:		op = LIR_restoreq; break;
#endif
		case LTy_F:		op = LIR_restoref; break;
		case LTy_D:		op = LIR_restored; break;
		default:		NanoAssert(0); break;
	}

	WRITE_OPCODE(op, _nIns, "%s %d", regNames[REGNUM(r)], d);
}

void Assembler::nSpill(const LIns* insToSpill) {
	Register			r			= insToSpill->getReg();
	int					d			= arDisp(insToSpill->getArIndex());

	underrunProtect(12);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);


	unsigned int op = 0;
	switch (insToSpill->retType()) {
		case LTy_I:		op = nanojit::LIR_spilli; break;
#if defined(NANOJIT_64BIT)
		case LTy_Q:		op = nanojit::LIR_spillq; break;
#endif
		case LTy_F:		op = nanojit::LIR_spillf; break;
		case LTy_D:		op = nanojit::LIR_spilld; break;
		default:		NanoAssert(0); break;
	}

	WRITE_OPCODE(op, _nIns, "%s %d", regNames[REGNUM(r)], d);
}

void Assembler::nSaveFlag(LIns* ins) {
	LIns*		oprnd1		= ins->oprnd1();
	LOpcode		op			= ins->opcode();
	Register	rr			= prepareResultReg(ins, GpRegs);

	// force value live
	findRegFor(oprnd1, GpRegs & ~rmask(rr));

	underrunProtect(8);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_OPCODE(op, _nIns, "%s", regNames[REGNUM(rr)]);

	freeResourcesOf(ins);
}

void Assembler::nRestoreFlag(LIns* ins) {
	LIns*		oprnd1		= ins->oprnd1();
	LOpcode		op			= ins->opcode();
	Register	rr			= findRegFor(oprnd1, GpRegs);

	underrunProtect(8);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_OPCODE(op, _nIns, "%s", regNames[REGNUM(rr)]);

	freeResourcesOf(ins);
}

void Assembler::asm_force_gen_flags(LIns* ins) {
	NanoAssert(ins->flagRequested() != 0);
	NanoAssert((ins->flagRequested() & (ArithFlag_Carry | ArithFlag_Overflow | ArithFlag_Borrow)) == 0);
	arm_arith_gen_flags(ins, ins->getReg(), ins->getReg(), ins->getReg());
}

void Assembler::nStoreToStack(LIns* ins, int d, RegisterMask allow) {

	Register r = findRegFor(ins, allow);

	unsigned int op = LIR_sentinel;

	switch (ins->retType()) {
		case LTy_I:
			op = LIR_pushi;
			break;

#if defined(NANOJIT_64BIT)
		case LTy_Q:
			op = LIR_pushq;
			break;
#endif

		case LTy_F:
			op = LIR_pushf;
			break;

		case LTy_D:
			op = LIR_pushd;
			break;

	}

	underrunProtect(12);
	WRITE_IMMI32(d, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %d", getRegName(r), d);
}

void Assembler::nGpCopy(Register d, Register s) {

	underrunProtect(12);
	WRITE_REGISTER_INDEX(s, _nIns);
	WRITE_REGISTER_INDEX(d, _nIns);
	WRITE_OPCODE(LIR_copyreg, _nIns, "%s %s", getRegName(d), getRegName(s));
}

void Assembler::nNonGpCopy(Register d, Register s) {
	nGpCopy(d, s);
}

void Assembler::nLoad32(LIns *ins) {
	LIns*			base	= ins->oprnd1();
	RegisterMask	allow	= GpRegs;
	Register		r		= prepareResultReg(ins, allow);
	Register		b		= findRegFor(base, allow & ~rmask(r));
	int				d		= ins->disp();

	underrunProtect(16);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(b, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s %s %d", getRegName(r), getRegName(b), d);

	freeResourcesOf(ins);

}

void Assembler::nLoad64(LIns *ins) {
	LIns*			base	= ins->oprnd1();
	RegisterMask	allow	= GpRegs;
	Register		r		= prepareResultReg(ins, allow);
	Register		b		= findRegFor(base, allow & ~rmask(r));
	int				d		= ins->disp();

	underrunProtect(16);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(b, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s %s %d", getRegName(r), getRegName(b), d);

	freeResourcesOf(ins);
}

void Assembler::nStore32(LOpcode op, LIns *value, int d, LIns *base) {

	Register b = findRegFor(base, GpRegs);
	Register r = findRegFor(value, GpRegs & ~rmask(b));

	underrunProtect(16);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(b, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %s %d", getRegName(r), getRegName(b), d);
}


void Assembler::nStore64(LOpcode op, LIns *value, int d, LIns *base) {
	Register b = findRegFor(base, GpRegs);
	Register r = findRegFor(value, GpRegs & ~rmask(b));

	underrunProtect(16);
	WRITE_IMMI32(static_cast<uint32_t>(d), _nIns);
	WRITE_REGISTER_INDEX(b, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %s %d", getRegName(r), getRegName(b), d);
}

void Assembler::nImmI(LIns *ins) {
	Register rd = prepareResultReg(ins, GpRegs);

	if (ins->getGenFlags() &&  (ins->flagRequested() != 0)) {
		NanoAssert((ins->flagRequested() & (ArithFlag_Overflow | ArithFlag_Carry | ArithFlag_Borrow)) == 0);
		arm_arith_gen_flags(ins, rd, rd, rd);
	}

	nImmI(rd, ins->immI());

	freeResourcesOf(ins);
}

#if defined(NANOJIT_64BIT)

void Assembler::nImmQ(LIns *ins) {
	Register rd = prepareResultReg(ins, GpRegs);

	if (ins->getGenFlags() &&  (ins->flagRequested() != 0)) {
		NanoAssert((ins->flagRequested() & (ArithFlag_Overflow | ArithFlag_Carry | ArithFlag_Borrow)) == 0);
		arm_arith_gen_flags(ins, rd, rd, rd);
	}

	nImmQ(rd, ins->immQ());

	freeResourcesOf(ins);
}

#endif

void Assembler::nImmF(LIns *ins) {
	Register rd = prepareResultReg(ins, GpRegs);

	nImmF(rd, ins->immF());

	freeResourcesOf(ins);
}

void Assembler::nImmD(LIns *ins) {
	Register rd = prepareResultReg(ins, GpRegs);

	nImmD(rd, ins->immD());

	freeResourcesOf(ins);
}

void Assembler::nCopyI(LIns* ins) {
	asm_op1(ins);
}

void Assembler::nCopyQ(LIns* ins) {
	asm_op1(ins);
}

void Assembler::nCopyF(LIns* ins) {
	asm_op1(ins);
}

void Assembler::nCopyD(LIns* ins) {
	asm_op1(ins);
}


void Assembler::nLabel() {
}

NIns* Assembler::nRelativeBranch(const NIns* target) {

	underrunProtect(sizeof(void*) + 4);
	ptrdiff_t jumpOffset = target - _nIns;
	WRITE_JUMP_OFFSET(jumpOffset, _nIns); // target must follow opcode for patching
	WRITE_OPCODE(LIR_branch, _nIns, "%p", target);

	return _nIns;
}

void Assembler::nRelativeBranchCond(bool onFalse, LIns *cond, const NIns *target, bool patchable, Branches* branches) {
	NanoUnusedVariable(patchable);

	NanoAssert(cond->oprnd1()->retType() == cond->oprnd2()->retType());
	NanoAssert(cond->oprnd1()->isLive());
	NanoAssert(cond->oprnd2()->isLive());

	LOpcode				lirCmpOp	= cond->opcode();
	LTy					lirCmpType	= cond->oprnd1()->retType();

	RegisterMask		allow		= GpRegs;
	Register			rca, rcb;

	rca = findRegFor(cond->oprnd1(), allow);
	allow &= ~rmask(rca);
	if (cond->oprnd1() != cond->oprnd2())
		rcb = findRegFor(cond->oprnd2(), allow);
	else
		rcb = rca;

	if (onFalse)
		lirCmpOp = invertCmpOpcode(lirCmpOp);

	underrunProtect(sizeof(void*) + 16);
	WRITE_REGISTER_INDEX(rcb, _nIns);
	WRITE_REGISTER_INDEX(rca, _nIns);
	WRITE_IMMI16(lirCmpType, _nIns);
	WRITE_IMMI16(lirCmpOp, _nIns);

	ptrdiff_t jumpOffset = target - _nIns;
	WRITE_JUMP_OFFSET(jumpOffset, _nIns); // target must follow opcode for patching

	WRITE_OPCODE(LIR_branch_cond, _nIns, "%s %s %s %p", getCmpOpName(lirCmpOp), getRegName(rca), getRegName(rcb), target);

	branches->addBranch(_nIns, target);
}

void Assembler::nCmov(LIns *ins) {
	RegisterMask		allow	= ins->isIorQ() ? GpRegs : FpRegs;
	LIns*				cond    = ins->oprnd1();
	LIns*				iftrue  = ins->oprnd2();
	LIns*				iffalse = ins->oprnd3();
	Register			rr, rf, rt, rca, rcb;


	NanoAssert(cond->oprnd1()->retType() == cond->oprnd2()->retType());
	setResultLive(cond->oprnd1());
	setResultLive(cond->oprnd2());

	NanoAssert(iftrue != iffalse);

	rr = prepareResultReg(ins, allow);
	allow &= ~rmask(rr);

	if (!iffalse->isInReg()) {
		rf = findRegFor(iffalse, allow);
		allow &= ~rmask(rf);
	} else {
		rf = iffalse->getReg();
	}

	if (!iftrue->isInReg()) {
		rt = findRegFor(iftrue, allow);
		allow &= ~rmask(rt);
	} else {
		rt = iftrue->getReg();
	}

	if (!cond->oprnd1()->isInReg()) {
		rca = findRegFor(cond->oprnd1(), allow);
		allow &= ~rmask(rca);
	} else {
		rca = cond->oprnd1()->getReg();
	}

	if (!cond->oprnd2()->isInReg()) {
		rcb = findRegFor(cond->oprnd2(), allow);
	} else {
		rcb = cond->oprnd2()->getReg();
	}

	LOpcode		lirCmpOp	= cond->opcode();
	LTy			lirCmpType	= cond->oprnd1()->retType();

	underrunProtect(28);
	WRITE_REGISTER_INDEX(rt, _nIns);
	WRITE_REGISTER_INDEX(rf, _nIns);
	WRITE_REGISTER_INDEX(rcb, _nIns);
	WRITE_REGISTER_INDEX(rca, _nIns);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_IMMI16(lirCmpType, _nIns);
	WRITE_IMMI16(lirCmpOp, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s <= (%s %s %s) ? %s %s", getRegName(rr), getRegName(rca), getCmpOpName(lirCmpOp), getRegName(rcb), getRegName(rt), getRegName(rf));

	freeResourcesOf(ins);
}

void Assembler::nCall(LIns* ins, const NIns *address) {

	LOpcode				op				= ins->opcode();
	ParameterValue*		callParams		= ins->callArgs();
	unsigned int		paramCount		= ins->callArgc();
	RegisterMask		evictMask		= ins->callVolatileMask();
	Register			r				= UnspecifiedReg;

	if (op != LIR_callv) {
		RegisterMask allow = ins->isIorQ() ? GpRegs : FpRegs;
		r = prepareResultReg(ins, allow);
		evictMask &= ~rmask(r);
	}

	evictSomeActiveRegs(evictMask);

	uint64_t signature = createFunctionSignature(ins, callParams, paramCount);

	underrunProtect(28);
	WRITE_IMMI64(signature , _nIns);
	WRITE_IMMI8(paramCount, _nIns);
	WRITE_IMMP(address, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %p %d", getRegName(r), address, paramCount);

	if (ins->opcode() != LIR_callv) {
		freeResourcesOf(ins);
	}

	reconcileParamRegs(callParams, paramCount);
}

// add addc sub subb and andc or xor mul (i and q types only)
void Assembler::nArith(LIns *ins) {
	LIns*			a			= ins->oprnd1();
	LIns*			b			= ins->oprnd2();
	RegisterMask	allow		= GpRegs;
	Register		rr, ra, rb;

	NanoAssert(ins->isLInsOp2());

	rr = prepareResultReg(ins, allow);
	allow &= ~rmask(rr);
	ra = findRegFor(a, allow);
	allow &= ~rmask(ra);
	if (a != b)
		rb = findRegFor(b, allow);
	else
		rb = ra;

	if (ins->getGenFlags() && (ins->flagRequested() != 0))
		arm_arith_gen_flags(ins, rr, ra, rb);
	
	underrunProtect(16);
	WRITE_REGISTER_INDEX(rb, _nIns);
	WRITE_REGISTER_INDEX(ra, _nIns);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s %s %s", getRegName(rr), getRegName(ra), getRegName(rb));

	freeResourcesOf(ins);
}

void Assembler::arm_arith_gen_flags(LIns* ins, Register rr, Register ra, Register rb) {

	uint32_t opcode = nanojit::LIR_genflgs;
	underrunProtect(24);
	WRITE_REGISTER_INDEX(rb, _nIns);
	WRITE_REGISTER_INDEX(ra, _nIns);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_IMMI16(static_cast<uint16_t>(ins->flagRequested()), _nIns);
	WRITE_IMMI16(static_cast<uint16_t>(ins->opcode()), _nIns);
	WRITE_OPCODE(opcode, _nIns, "%02x %s %s %s %s", ins->flagRequested(), interpreter::getInterpreterOpcodeString(ins->opcode()), getRegName(rr), getRegName(ra), getRegName(rb));
}

// i and q types only
void Assembler::nNegNot(LIns* ins) {
	LOpcode		op	= ins->opcode();
	Register	rd	= prepareResultReg(ins, GpRegs);
	Register	rs	= findRegFor(ins->oprnd1(), GpRegs & ~rmask(rd));

	NanoAssert(ins->isLInsOp1());

	if (ins->getGenFlags() && (ins->flagRequested() != 0)) {
		NanoAssert((ins->flagRequested() & (ArithFlag_Carry | ArithFlag_Overflow | ArithFlag_Borrow)) == 0);
		arm_arith_gen_flags(ins, rd, rs, rs);
	}

	underrunProtect(16);
	WRITE_REGISTER_INDEX(rs, _nIns);
	WRITE_REGISTER_INDEX(rd, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %s", getRegName(rd), getRegName(rs));

	freeResourcesOf(ins);
}

// add sub div mul (f, d, f4 types)
void Assembler::nFop(LIns *ins) {
	nArith(ins);
}

// f, d, f4 types
void Assembler::nNegAbs(LIns *ins) {
	nNegNot(ins);
}

void Assembler::nMull(LIns* ins) {

	// value is live, there was no used of the hi value
	// fallback on a mul instruction

	LOpcode     op    = ins->opcode();
	LIns*       a     = ins->oprnd1();
	LIns*       b     = ins->oprnd2();

	LOpcode		mulOp = LIR_sentinel;

	switch (op) {
		case LIR_mullsi:
		case LIR_mullui:
			mulOp = LIR_muli;
			break;
#if defined(NANOJIT_64BIT)
		case LIR_mullsq:
		case LIR_mulluq:
			mulOp = LIR_mulq;
			break;
#endif

		default:
			NanoAssert(0);
			break;
	}

	RegisterMask allow = GpRegs;
	Register rd, ra, rb;

	rd = prepareResultReg(ins, allow);
	allow &= ~rmask(rd);
	ra = findRegFor(a, allow);
	allow &= ~rmask(ra);
	if (a != b)
		rb = findRegFor(b, allow);
	else
		rb = ra;


	underrunProtect(20);
	WRITE_REGISTER_INDEX(rb, _nIns);
	WRITE_REGISTER_INDEX(ra, _nIns);
	WRITE_REGISTER_INDEX(rd, _nIns);
	WRITE_OPCODE(mulOp, _nIns, "%s %s %s", getRegName(rd), getRegName(ra), getRegName(rb));

	freeResourcesOf(ins);
}

void Assembler::nMullhi(LIns* ins) {
	LIns*       insLo = ins->oprnd1();
	LIns*       insHi = ins;
	LOpcode     op    = insLo->opcode();
	LIns*       a     = insLo->oprnd1();
	LIns*       b     = insLo->oprnd2();

	setResultLive(a);
	setResultLive(b);

	Register rdlo, rdhi, ra, rb;
	RegisterMask allow = GpRegs;

	rdhi = prepareResultReg(insHi, allow);
	allow &= ~rmask(rdhi);

	rdlo = prepareResultReg(insLo, allow);
	allow &= ~rmask(rdlo);

	ra = findRegFor(a, allow);
	allow &= ~rmask(ra);

	if (a != b)
		rb = findRegFor(b, allow);
	else
		rb = ra;

	if (insLo->getGenFlags() && (insLo->flagRequested() != 0))
		asm_force_gen_flags(insLo);
	if (insHi->getGenFlags() && (insHi->flagRequested() != 0))
		asm_force_gen_flags(insHi);

	underrunProtect(20);
	WRITE_REGISTER_INDEX(rb, _nIns);
	WRITE_REGISTER_INDEX(ra, _nIns);
	WRITE_REGISTER_INDEX(rdhi, _nIns);
	WRITE_REGISTER_INDEX(rdlo, _nIns);
	WRITE_OPCODE(op, _nIns, "%s %s = %s * %s", getRegName(rdlo), getRegName(rdhi), getRegName(ra), getRegName(rb));

	freeResourcesOf(insLo);
	freeResourcesOf(insHi);
}

void Assembler::nRecipSqrt(LIns* ins) {
	asm_op1(ins);
}


// following are conversions / reinterpretations.


void Assembler::nMovI(LIns* ins) {
	asm_op1(ins);
}

#if defined(NANOJIT_64BIT)
void Assembler::nMovQ(LIns* ins) {
	asm_op1(ins);
}
#endif

void Assembler::nMovF(LIns* ins) {
	asm_op1(ins);
}

void Assembler::nMovD(LIns* ins) {
	asm_op1(ins);
}


void Assembler::nFconv(LIns* ins) {
	asm_op1(ins);
}



uint64_t Assembler::createFunctionSignature(LIns* callIns, const ParameterValue* callParams, unsigned int paramCount) {

	LTy				type;
	uint64_t		signature = 0;
	unsigned int	shiftCount = 3;

	// 3 bits for each type, max 20 parameters + return value
	NanoAssert((paramCount + 1) < ((sizeof(uint64_t) * CHAR_BIT) / 3));

	type = callIns->retType();
	NanoAssert(type < 7);
	signature = (type + 1);

	for (unsigned int paramIndex = 0; paramIndex < paramCount; ++paramIndex) {
		const ParameterValue* param = &callParams[paramIndex];
		LTy type = param->value->retType();
		NanoAssert(type < 8);
		signature |= static_cast<uint64_t>(type + 1) << shiftCount;

		shiftCount += 3;
	}

	return signature;
}



void Assembler::asm_restore_remat(LIns *ins, Register r) {
	findRegFor(ins, rmask(r));

	if (ins->isImmI())
		nImmI(r, ins->immI());
#if defined(NANOJIT_64BIT)
	else if (ins->isImmQ())
		nImmQ(r, ins->immQ());
#endif
	else if (ins->isImmF())
		nImmF(r, ins->immF());
	else if (ins->isImmD())
		nImmD(r, ins->immD());
	else
		NanoAssert(0); // invalid opcode ?
}

void Assembler::nImmI(Register r, uint32_t v) {
	underrunProtect(12);
	WRITE_IMMI32(v, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(LIR_immi, _nIns, "%s %d(%08x)", getRegName(r), v, v);
}


#if defined(NANOJIT_64BIT)

void Assembler::nImmQ(Register r, uint64_t v) {
	underrunProtect(16);
	WRITE_IMMI64(v, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(LIR_immq, _nIns, "%s %lld(%016x)", getRegName(r), v, v);
}

#endif

void Assembler::nImmF(Register r, float v) {
	underrunProtect(12);
	WRITE_IMMF32(v, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(LIR_immf, _nIns, "%s %f", getRegName(r), v);
}

void Assembler::nImmD(Register r, double v) {
	underrunProtect(16);
	WRITE_IMMF64(v, _nIns);
	WRITE_REGISTER_INDEX(r, _nIns);
	WRITE_OPCODE(LIR_immd, _nIns, "%s %f", getRegName(r), v);
}

void Assembler::asm_op1(LIns* ins) {

	NanoAssert(ins->isLInsOp1());
	
	Register rd		= prepareResultReg(ins, GpRegs);
	Register rs		= findRegFor(ins->oprnd1(), GpRegs & ~rmask(rd));

	if (ins->getGenFlags() && ins->flagRequested()) {
		asm_force_gen_flags(ins);
	}

	underrunProtect(12);
	WRITE_REGISTER_INDEX(rs, _nIns);
	WRITE_REGISTER_INDEX(rd, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s %s", getRegName(rd), getRegName(rs));

	freeResourcesOf(ins);
}

void Assembler::asm_op2(LIns* ins) {

	NanoAssert(ins->isLInsOp2());

	LIns*			a			= ins->oprnd1();
	LIns*			b			= ins->oprnd2();
	RegisterMask	allow		= GpRegs;
	Register		rr, ra, rb;

	rr = prepareResultReg(ins, allow);
	allow &= ~rmask(rr);
	ra = findRegFor(a, allow);
	allow &= ~rmask(ra);
	if (a != b)
		rb = findRegFor(b, allow);
	else
		rb = ra;

	if (ins->getGenFlags() && ins->flagRequested()) {
		asm_force_gen_flags(ins);
	}

	underrunProtect(16);
	WRITE_REGISTER_INDEX(rb, _nIns);
	WRITE_REGISTER_INDEX(ra, _nIns);
	WRITE_REGISTER_INDEX(rr, _nIns);
	WRITE_OPCODE(ins->opcode(), _nIns, "%s %s %s", getRegName(rr), getRegName(ra), getRegName(rb));

	freeResourcesOf(ins);
}



} // namespace nanojit

