
#include "NativeInterpreterVmCallbacks.h"
#include "math.h"

namespace nanojit
{
namespace interpreter
{


const uint8_t* callbackGenFlags(IntepreterContext* context, const uint8_t* stream) {

	NanoUnusedVariable(context);
	unsigned int	arithOp;
	uint32_t		flagRequested;
	uint32_t		rr, ra, rb;

	READ_IMMI16(arithOp, stream);
	READ_IMMI16(flagRequested, stream);
	READ_REGISTER_INDEX(rr, stream);
	READ_REGISTER_INDEX(ra, stream);
	READ_REGISTER_INDEX(rb, stream);


	uint32_t valueResult	= context->get<uint32_t>(rr);
	uint32_t valueOp1		= context->get<uint32_t>(ra);
	uint32_t valueOp2		= context->get<uint32_t>(rb);

	if (arithOp == LIR_copyi) {
		valueOp1 = valueOp1;
	}

	if ((flagRequested & ArithFlag_Zero) != 0) {
		if (valueResult == 0)
			context->setFlag(ArithFlag_Zero);
		else
			context->clearFlag(ArithFlag_Zero);
	}

	if ((flagRequested & ArithFlag_Sign) != 0) {
		if ((valueResult >> 31) != 0)
			context->setFlag(ArithFlag_Sign);
		else
			context->clearFlag(ArithFlag_Sign);
	}

	if ((flagRequested & ArithFlag_Carry) != 0) {
		if (
			(arithOp == LIR_addi) || (arithOp == LIR_addci)
#if defined(NANOJIT_64BIT)
			||	(arithOp == LIR_addq) || (arithOp == LIR_addcq)
#endif
			) {
				uint32_t carry = ((valueOp1 & valueOp2) | ((valueOp1 | valueOp2) & ~valueResult)) >> 31;
				if (carry != 0)
					context->setFlag(ArithFlag_Carry);
				else
					context->clearFlag(ArithFlag_Carry);
		} else if (
			(arithOp == LIR_lshi)
#if defined(NANOJIT_64BIT)
			||	(arithOp == LIR_lshq)
#endif
			) {
				uint32_t carry = (valueOp1 >> (32 - valueOp2)) & 0x01;
				if ((carry != 0) && (valueOp2 != 0))
					context->setFlag(ArithFlag_Carry);
				else
					context->clearFlag(ArithFlag_Carry);
		} else if (
			(arithOp == LIR_rshi) || (arithOp == LIR_rshui) || (arithOp == LIR_rori) || (arithOp == LIR_rcri)
#if defined(NANOJIT_64BIT)
			||	(arithOp == LIR_rshuq) || (arithOp == LIR_rshq) || (arithOp == LIR_rorq) || (arithOp == LIR_rcrq)
#endif
			) {
				uint32_t carry = (valueOp1 >> (valueOp2 - 1)) & 0x01;
				if ((carry != 0) && (valueOp2 != 0))
					context->setFlag(ArithFlag_Carry);
				else
					context->clearFlag(ArithFlag_Carry);
		} else {
			NanoAssert(0); // unsupported opcode
		}

	}

	if ((flagRequested & ArithFlag_Borrow) != 0) {
#if defined(NANOJIT_64BIT)
		NanoAssert((arithOp == LIR_subi) || (arithOp == LIR_subbi) || (arithOp == LIR_subq) || (arithOp == LIR_subbq));
#else
		NanoAssert((arithOp == LIR_subi) || (arithOp == LIR_subbi));
#endif

		uint32_t borrow = ((~valueOp1 & valueOp2) | ((~valueOp1 | valueOp2) & valueResult)) >> 31;
		if (borrow != 0)
			context->setFlag(ArithFlag_Borrow);
		else
			context->clearFlag(ArithFlag_Borrow);
	}

	if ((flagRequested & ArithFlag_Overflow) != 0) {
		uint32_t overflowTest = 0;
		if (
			(arithOp == LIR_addi) || (arithOp == LIR_addci) 
#if defined(NANOJIT_64BIT)
			||	(arithOp == LIR_addq) || (arithOp == LIR_addcq)
#endif
			) {
				overflowTest = (valueOp1 ^ valueResult) & (valueOp2 ^ valueResult);
		} else if (
			(arithOp == LIR_subi) || (arithOp == LIR_subbi)
#if defined(NANOJIT_64BIT)
			||	(arithOp == LIR_subq) || (arithOp == LIR_subbq)
#endif
			) {
				overflowTest = (valueOp1 ^ valueResult) & ~(valueOp2 ^ valueResult);
		} else  {
			NanoAssert(0); // unsupported opcode
		}

		uint32_t overflow = (overflowTest >> 31) & 0x01;
		if (overflow != 0)
			context->setFlag(ArithFlag_Overflow);
		else
			context->clearFlag(ArithFlag_Overflow);
	}

	return stream;
}



template<typename T, typename TS>
bool doTestImpl(IntepreterContext* context, unsigned int ra, unsigned int rb, LOpcode op) {

	bool		result		= false;
	T			va			= context->get<T>(ra);
	T			vb			= context->get<T>(rb);

	switch (op) {
	case LIR_eqi:
#if defined(NANOJIT_64BIT)
	case LIR_eqq:
#endif
	case LIR_eqf:
	case LIR_eqd:
		result = (va == vb); 
		break;

	case LIR_nei:
#if defined(NANOJIT_64BIT)
	case LIR_neq:
#endif
	case LIR_nef:
	case LIR_ned:
		result = (va != vb); 
		break;

	case LIR_lti:
#if defined(NANOJIT_64BIT)
	case LIR_ltq:
#endif
	case LIR_ltf:
	case LIR_ltd:
		result = (static_cast<TS>(va) < static_cast<TS>(vb)); 
		break;


	case LIR_gei:
#if defined(NANOJIT_64BIT)
	case LIR_geq:
#endif
	case LIR_gef:
	case LIR_ged:
		result = (static_cast<TS>(va) >= static_cast<TS>(vb)); 
		break;

	case LIR_gti:
#if defined(NANOJIT_64BIT)
	case LIR_gtq:
#endif
	case LIR_gtf:
	case LIR_gtd:
		result = (static_cast<TS>(va) > static_cast<TS>(vb)); 
		break;

	case LIR_lei:
#if defined(NANOJIT_64BIT)
	case LIR_leq:
#endif
	case LIR_lef:
	case LIR_led:
		result = (static_cast<TS>(va) <= static_cast<TS>(vb)); 
		break;

	case LIR_ltui:
#if defined(NANOJIT_64BIT)
	case LIR_ltuq:
#endif
		result = (va < vb); 
		break;

	case LIR_geui:
#if defined(NANOJIT_64BIT)
	case LIR_geuq:
#endif
		result = (va >= vb); 
		break;

	case LIR_gtui:
#if defined(NANOJIT_64BIT)
	case LIR_gtuq:
#endif
		result = (va > vb); 
		break;

	case LIR_leui:
#if defined(NANOJIT_64BIT)
	case LIR_leuq:
#endif
		result = (va <= vb); 
		break;

	case LIR_unf:
	case LIR_und:
		result = (va != va) || (vb != vb);
		break;

	default:
		NanoAssert(0);
		break;
	}

	return result;
}

bool doTest(IntepreterContext* context, unsigned int ra, unsigned int rb, LOpcode op, LTy type) {

	bool conditionResult = false;

	switch (type) {
		case LTy_I:
			conditionResult = doTestImpl<uint32_t, int32_t>(context, ra, rb, op);
			break;
		case LTy_Q:
			conditionResult = doTestImpl<uint64_t, int64_t>(context, ra, rb, op);
			break;

		case LTy_F:
			conditionResult = doTestImpl<float32_t, float32_t>(context, ra, rb, op);
			break;

		case LTy_D:
			conditionResult = doTestImpl<float64_t, float64_t>(context, ra, rb, op);
			break;

		default:
			NanoAssert(0);
	}

	return conditionResult;
}




/*
const uint8_t* callback(IntepreterContext* context, const uint8_t* stream) {
	uint32_t rd, rs;
	READ_REGISTER_INDEX(rd, stream);
	READ_REGISTER_INDEX(rs, stream);

	return stream;
}
*/


} // namespace interpreter
} // namespace nanojit


