/* -*- Mode: C++; c-basic-offset: 4; indent-tabs-mode: nil; tab-width: 4 -*- */
/* vi: set ts=4 sw=4 expandtab: (add to ~/.vimrc: set modeline modelines=5) */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nanojit.h"

// uncomment this to enable _vprof/_nvprof macros
//#define DOPROF
#include "vprof/vprof.h"

#if defined(NANOJIT_HOST_PLATFORM_X64)

/*
completion
- 64bit branch offsets
- finish cmov/qcmov with other conditions

better code
- put R12 back in play as a base register
- no-disp addr modes (except RBP/R13)
- disp64 branch/call
- spill gp values to xmm registers?
- prefer xmm registers for copies since gprs are in higher demand?
- stack based LIR_paramp

tracing
- nFragExit

*/



namespace nanojit
{

	const char *regNames[] = {
		"rax",  "rcx",  "rdx",   "rbx",   "rsp",   "rbp",   "rsi",   "rdi",
		"r8",   "r9",   "r10",   "r11",   "r12",   "r13",   "r14",   "r15",
		"xmm0", "xmm1", "xmm2",  "xmm3",  "xmm4",  "xmm5",  "xmm6",  "xmm7",
		"xmm8", "xmm9", "xmm10", "xmm11", "xmm12", "xmm13", "xmm14", "xmm15"
	};

	const char *gpRegNames32[] = {
		"rax(d)", "rcx(d)", "rdx(d)", "rbx(d)", "rsp(d)", "rbp(d)", "rsi(d)", "rdi(d)",
		"r8(d)", "r9(d)", "r10(d)", "r11(d)", "r12(d)", "r13(d)", "r14(d)", "r15(d)"
	};

	const char *gpRegNames8[] = {
		"rax(l)", "rcx(l)", "rdx(l)", "rbx(l)", "rsp(l)", "rbp(l)", "rsi(l)", "rdi(l)",
		"r8(l)", "r9(l)", "r10(l)", "r11(l)", "r12(l)", "r13(l)", "r14(l)", "r15(l)"
	};

	const char *gpRegNames8hi[] = {
		"rax(h)", "rcx(h)", "rdx(h)", "rbx(h)"
	};

	const char *gpRegNames16[] = {
		"rax(w)", "rcx(w)", "rdx(w)", "bx", "rsp(w)", "rbp(w)", "rsi(w)", "rdi(w)",
		"r8(w)", "r9(w)", "r10(w)", "r11(w)", "r12(w)", "r13(w)", "r14(w)", "r15(w)"
	};

#ifdef NANOJIT_DEBUG
	#define TODO(x) todo(#x)
	static void todo(const char *s) {
		NanoLog("%s",s);
		NanoAssertMsgf(false, "%s", s);
	}
#else
	#define TODO(x)
#endif


	// MODRM and SIB restrictions:
	// memory access modes != 11 require SIB if base&7 == 4 (RSP or R12)
	// mode 00 with base == x101 means RIP+disp32 (RBP or R13), use mode 01 disp8=0 instead
	// mode 01 or 11 with base = x101 means disp32 + EBP or R13, not RIP relative
	// base == x100 means SIB byte is present, so using ESP|R12 as base requires SIB
	// rex prefix required to use RSP-R15 as 8bit registers in mod/rm8 modes.

	// take R12 and RSP out of play as a base register because using ESP or R12 as base requires the SIB byte
	const RegisterMask BaseRegs = GpRegs & ~rmask(R12) & ~rmask(RSP);

	static inline int oplen(uint64_t op) {
		return op & 255;
	}

	// encode 2-register rex prefix.  dropped if none of its bits are set.
	static inline uint64_t rexrb(uint64_t op, Register r, Register b) {
		int shift = 64 - 8*oplen(op);
		NanoAssert( (((op>>shift)&255) == 0x40) ||
					(((op>>shift)&255) == 0x48) ); // Make sure rexrb is properly used
		uint64_t rex = ((op >> shift) & 255) | ((REGNUM(r)&8)>>1) | ((REGNUM(b)&8)>>3);
		return rex != 0x40 ? op | rex << shift : op - 1;
	}

	// encode 3-register rex prefix.  dropped if none of its bits are set.
	static inline uint64_t rexrxb(uint64_t op, Register r, Register x, Register b) {
		int shift = 64 - 8*oplen(op);
		uint64_t rex = ((op >> shift) & 255) | ((REGNUM(r)&8)>>1) | ((REGNUM(x)&8)>>2) | ((REGNUM(b)&8)>>3);
		return rex != 0x40 ? op | rex << shift : op - 1;
	}

	// encode 2-register rex prefix.  dropped if none of its bits are set, but
	// keep REX if b >= rsp, to allow uniform use of all 16 8bit registers
	static inline uint64_t rexrb8(uint64_t op, Register r, Register b) {
		int shift = 64 - 8*oplen(op);
		uint64_t rex = ((op >> shift) & 255) | ((REGNUM(r)&8)>>1) | ((REGNUM(b)&8)>>3);
		return ((rex | (REGNUM(b) & ~3)) != 0x40) ? (op | (rex << shift)) : op - 1;
	}

	// encode 2-register rex prefix that follows a manditory prefix (66,F2,F3)
	// [prefix][rex][opcode]
	static inline uint64_t rexprb(uint64_t op, Register r, Register b) {
		int shift = 64 - 8*oplen(op) + 8;
#ifdef NANOJIT_DEBUG
		int mandatory_prefix = (op >> (shift-8)) & 255;
		NanoAssert(mandatory_prefix!=0);
		NanoAssert(mandatory_prefix==0x66 || mandatory_prefix == 0xF2
				   || mandatory_prefix == 0xF3 );
#endif
		uint64_t rex = ((op >> shift) & 255) | ((REGNUM(r)&8)>>1) | ((REGNUM(b)&8)>>3);
		// to drop rex, we replace rex with manditory prefix, and decrement length
		return rex != 0x40 ? op | rex << shift :
			((op & ~(255LL<<shift)) | (op>>(shift-8)&255) << shift) - 1;
	}

	// [rex][opcode][mod-rr]
	static inline uint64_t mod_rr(uint64_t op, Register r, Register b) {
		return op | uint64_t((REGNUM(r)&7)<<3 | (REGNUM(b)&7))<<56;
	}

	// [rex][opcode][modrm=r][sib=xb]
	static inline uint64_t mod_rxb(uint64_t op, Register r, Register x, Register b) {
		return op | /*modrm*/uint64_t((REGNUM(r)&7)<<3)<<48 | /*sib*/uint64_t((REGNUM(x)&7)<<3|(REGNUM(b)&7))<<56;
	}

	static inline uint64_t mod_disp32(uint64_t op, Register r, Register b, int32_t d, int allow_8) {
		NanoAssert(IsGpReg(r) && IsGpReg(b));
		NanoAssert((REGNUM(b) & 7) != 4); // using RSP or R12 as base requires SIB
		uint64_t mod = (((op>>24)&255)>>6); // mod bits in addressing mode: 0,1,2, or 3
		if (allow_8 && (mod == 2) && isS8(d)) {
			// op is:  0x[disp32=0][mod=2:r:b][op][rex][len]
			int len = oplen(op);
			op = (op & ~0xff000000LL) | (0x40 | (REGNUM(r)&7)<<3 | (REGNUM(b)&7))<<24; // replace mod
			return op<<24 | int64_t(d)<<56 | (len-3); // shrink disp, add disp8
		} else {
			// op is: 0x[disp32][mod][op][rex][len]
			return op | int64_t(d)<<32 | uint64_t((REGNUM(r)&7)<<3 | (REGNUM(b)&7))<<24;
		}
	}

	// All the emit() functions should only be called from within codegen
	// functions PUSHR(), SHR(), etc.

	void Assembler::emit(uint64_t op) {
		int len = oplen(op);
		// we will only move nIns by -len bytes, but we write 8
		// bytes. so need to protect 8 so we dont stomp the page
		// header or the end of the preceding page (might segf)
		underrunProtect(8);
		((int64_t*)_nIns)[-1] = op;
		_nIns -= len; // move pointer by length encoded in opcode
		_nvprof("x64-bytes", len);
	}

	void Assembler::emit8(uint64_t op, int64_t v) {
		NanoAssert(isS8(v));
		emit(op | uint64_t(v)<<56);
	}

	void Assembler::emit_target8(size_t underrun, uint64_t op, const NIns* target) {
		underrunProtect(underrun); // must do this before calculating offset
		// Nb: see emit_target32() for why we use _nIns here.
		int64_t offset = target - _nIns;
		NanoAssert(isS8(offset));
		emit(op | uint64_t(offset)<<56);
	}

	void Assembler::emit_target32(size_t underrun, uint64_t op, const NIns* target) {
		underrunProtect(underrun); // must do this before calculating offset
		// Nb: at this point in time, _nIns points to the most recently
		// written instruction, ie. the jump's successor.  So why do we use it
		// to compute the offset, rather than the jump's address?  Because in
		// x86/x64-64 the offset in a relative jump is not from the jmp itself
		// but from the following instruction.  Eg. 'jmp $0' will jump to the
		// next instruction.
		int64_t offset = target ? target - _nIns : 0;
		if (!isS32(offset))
			setError(BranchTooFar);
		emit(op | uint64_t(uint32_t(offset))<<32);
	}

	void Assembler::emit_targetIpRel(size_t underrun, uint64_t op, uint32_t offset) {
		NanoAssert(underrun >= 16);
		underrunProtect(underrun); // must do this before calculating offset
		// Nb: at this point in time, _nIns points to the most recently
		// written instruction, ie. the jump's successor.
		_nIns -= 4;
		*((uint32_t*)_nIns) = offset;
		emit(op);
	}

	// 3-register modrm32+sib form
	void Assembler::emitrxb(uint64_t op, Register r, Register x, Register b) {
		emit(rexrxb(mod_rxb(op, r, x, b), r, x, b));
	}

	// 2-register modrm32 form
	void Assembler::emitrr(uint64_t op, Register r, Register b) {
		emit(rexrb(mod_rr(op, r, b), r, b));
	}

	// 2-register modrm8 form (8 bit operand size)
	void Assembler::emitrr8(uint64_t op, Register r, Register b) {
		emit(rexrb8(mod_rr(op, r, b), r, b));
	}

	// same as emitrr, but with a prefix byte
	void Assembler::emitprr(uint64_t op, Register r, Register b) {
		emit(rexprb(mod_rr(op, r, b), r, b));
	}

	// disp32 modrm8 form, when the disp fits in the instruction (opcode is 1-3 bytes)
	void Assembler::emitrm8(uint64_t op, Register r, int32_t d, Register b) {
		emit(rexrb8(mod_disp32(op, r, b, d, 1), r, b));
	}

	// disp32 modrm form, when the disp fits in the instruction (opcode is 1-3 bytes)
	void Assembler::emitrm32(uint64_t op, Register r, int32_t d, Register b) {
		emit(rexrb(mod_disp32(op, r, b, d, 0), r, b));
	}

	// disp32 modrm form, when the disp fits in the instruction (opcode is 1-3 bytes)
	void Assembler::emitrm(uint64_t op, Register r, int32_t d, Register b) {
		emit(rexrb(mod_disp32(op, r, b, d, 1), r, b));
	}

	// disp32 modrm form when the disp must be written separately (opcode is 4+ bytes)
#define emit_disp32_sib(o,d) (emit_disp32( ((o)&0x00FFFFFFFFFFFFFFLL)<<8, d, 1) >>8) | ((o)&0xFF00000000000000LL)
	uint64_t Assembler::emit_disp32(uint64_t op, int32_t d, int allow_8)    {
		if (allow_8 && isS8(d)) {
			NanoAssert(((op>>56)&0xC0) == 0x80); // make sure mod bits == 2 == disp32 mode
			underrunProtect(1+8);
			*(--_nIns) = (NIns) d;
			_nvprof("x64-bytes", 1);
			op ^= 0xC000000000000000LL; // change mod bits to 1 == disp8 mode
		} else {
			underrunProtect(4+8); // room for displ plus fullsize op
			*((int32_t*)(_nIns -= 4)) = d;
			_nvprof("x64-bytes", 4);
		}
		return op;
	}

	// disp32 modrm form when the disp must be written separately (opcode is 4+ bytes)
	void Assembler::emitrm_wide(uint64_t op, Register r, int32_t d, Register b) {
		op = emit_disp32(op, d, 1);
		emitrr(op, r, b);
	}

	// disp32 modrm form when the disp must be written separately (opcode is 4+ bytes)
	void Assembler::emitrm_wide32(uint64_t op, Register r, int32_t d, Register b) {
		op = emit_disp32(op, d, 0);
		emitrr(op, r, b);
	}

	// disp32 modrm form when the disp must be written separately (opcode is 4+ bytes)
	// p = prefix -- opcode must have a 66, F2, or F3 prefix
	void Assembler::emitprm(uint64_t op, Register r, int32_t d, Register b) {
		op = emit_disp32(op, d, 1);
		emitprr(op, r, b);
	}

	// disp32 modrm form when the disp must be written separately (opcode is 4+ bytes)
	// p = prefix -- opcode must have a 66, F2, or F3 prefix
	void Assembler::emitprm32(uint64_t op, Register r, int32_t d,   Register b) {
		op = emit_disp32(op, d, 0);
		emitprr(op, r, b);
	}


	// disp32 modrm form with 32-bit immediate value
	void Assembler::emitrm_imm32(uint64_t op, Register b, int32_t d, int32_t imm) {
		NanoAssert(IsGpReg(b));
		NanoAssert((REGNUM(b) & 7) != 4); // using RSP or R12 as base requires SIB
		underrunProtect(4+4+8); // room for imm plus disp plus fullsize op
		*((int32_t*)(_nIns -= 4)) = imm;
		_nvprof("x86-bytes", 4);
		emitrm_wide(op, RZero, d, b);
	}

	// disp32 modrm form with 16-bit immediate value
	// p = prefix -- opcode must have a 66, F2, or F3 prefix
	void Assembler::emitprm_imm16(uint64_t op, Register b, int32_t d, int32_t imm) {
		NanoAssert(IsGpReg(b));
		NanoAssert((REGNUM(b) & 7) != 4); // using RSP or R12 as base requires SIB
		underrunProtect(2+4+8); // room for imm plus disp plus fullsize op
		*((int16_t*)(_nIns -= 2)) = (int16_t) imm;
		_nvprof("x86-bytes", 2);
		emitprm(op, RZero, d, b);
	}

	// disp32 modrm form with 8-bit immediate value
	void Assembler::emitrm_imm8(uint64_t op, Register b, int32_t d, int32_t imm) {
		NanoAssert(IsGpReg(b));
		NanoAssert((REGNUM(b) & 7) != 4); // using RSP or R12 as base requires SIB
		underrunProtect(1+4+8); // room for imm plus disp plus fullsize op
		*((int8_t*)(_nIns -= 1)) = (int8_t) imm;
		_nvprof("x86-bytes", 1);
		emitrm_wide(op, RZero, d, b);
	}

	void Assembler::emitrr_imm(uint64_t op, Register r, Register b, int32_t imm) {
		NanoAssert(IsGpReg(r) && IsGpReg(b));
		underrunProtect(4+8); // room for imm plus fullsize op
		*((int32_t*)(_nIns -= 4)) = imm;
		_nvprof("x86-bytes", 4);
		emitrr(op, r, b);
	}

	void Assembler::emitrr_imm8(uint64_t op, Register r, Register b, uint8_t imm) {
		NanoAssert(IsFpReg(r) && IsFpReg(b));
		underrunProtect(1+8); // room for imm plus fullsize op
		*((uint8_t*)(_nIns -= 1)) = imm;
		_nvprof("x86-bytes", 1);
		emitrr(op, r, b);
	}

	void Assembler::emitprr_imm8(uint64_t op, Register r, Register b, uint8_t imm) {
		NanoAssert((IsGpReg(r) && IsGpReg(b)) || (IsFpReg(r) && IsFpReg(b)));
		underrunProtect(1+8); // room for imm plus fullsize op
		*((uint8_t*)(_nIns -= 1)) = imm;
		_nvprof("x86-bytes", 1);
		emitprr(op, r, b);
	}

	void Assembler::emitr_imm64(uint64_t op, Register r, uint64_t imm64) {
		underrunProtect(8+8); // imm64 + worst case instr len
		*((uint64_t*)(_nIns -= 8)) = imm64;
		_nvprof("x64-bytes", 8);
		emitr(op, r);
	}

	void Assembler::emitrxb_imm(uint64_t op, Register r, Register x, Register b, int32_t imm) {
		NanoAssert(IsGpReg(r) && IsGpReg(x) && IsGpReg(b));
		underrunProtect(4+8); // room for imm plus fullsize op
		*((int32_t*)(_nIns -= 4)) = imm;
		_nvprof("x86-bytes", 4);
		emitrxb(op, r, x, b);
	}

	// op = [rex][opcode][modrm][imm8]
	void Assembler::emitr_imm8(uint64_t op, Register b, int32_t imm8) {
		NanoAssert(IsGpReg(b) && isS8(imm8));
		op |= uint64_t(imm8)<<56 | uint64_t(REGNUM(b)&7)<<48;  // modrm is 2nd to last byte
		emit(rexrb(op, RZero, b));
	}


	void Assembler::emitxm_abs(uint64_t op, Register r, int32_t addr32)
	{
		underrunProtect(4+8);
		*((int32_t*)(_nIns -= 4)) = addr32;
		_nvprof("x64-bytes", 4);
		op = op | uint64_t((REGNUM(r)&7)<<3)<<48; // put rr[0:2] into mod/rm byte
		op = rexrb(op, r, RZero);         // put rr[3] into rex byte
		emit(op);
	}

	void Assembler::emitxm_rel(uint64_t op, Register r, NIns* addr64)
	{
		underrunProtect(4+8);
		int32_t d = (int32_t)(addr64 - _nIns);
		*((int32_t*)(_nIns -= 4)) = d;
		_nvprof("x64-bytes", 4);
		emitrr(op, r, RZero);
	}

	// Succeeds if 'target' is within a signed 8-bit offset from the current
	// instruction's address.
	bool Assembler::isTargetWithinS8(const NIns* target)
	{
		NanoAssert(target);
		// First call underrunProtect().  Without it, we might compute the
		// difference just before starting a new code chunk.
		underrunProtect(8);
		return isS8(target - _nIns);
	}

	// Like isTargetWithinS8(), but for signed 32-bit offsets.
	bool Assembler::isTargetWithinS32(const NIns* target,int32_t maxInstSize)
	{
		NanoAssert(target);
		// some instructions with S32 offsets take more than 8 bytes (e.g. packed float loads like movaps/movups)
		underrunProtect(maxInstSize);
		return isS32(target - _nIns);
	}

#define RB(r)       gpRegNames8[(REGNUM(r))]
#define RS(r)       gpRegNames16[(REGNUM(r))]
#define RBhi(r)     gpRegNames8hi[(REGNUM(r))]
#define RL(r)       gpRegNames32[(REGNUM(r))]
#define RQ(r)       gpn(r)

	typedef Register R;
	typedef int      I;
	typedef int32_t  I32;
	typedef uint64_t U64;
	typedef size_t   S;

	void Assembler::PUSHR(R r)  { emitr(X64_pushr,r); asm_output("push %s", RQ(r)); }
	void Assembler::POPR( R r)  { emitr(X64_popr, r); asm_output("pop %s", RQ(r)); }
	void Assembler::NOT(  R r)  { emitr(X64_not,  r); asm_output("notl %s", RL(r)); }
	void Assembler::NEG(  R r)  { emitr(X64_neg,  r); asm_output("negl %s", RL(r)); }
	void Assembler::NOTQ( R r)  { emitr(X64_notq,  r); asm_output("notq %s", RL(r)); }
	void Assembler::NEGQ( R r)  { emitr(X64_negq,  r); asm_output("negq %s", RL(r)); }
	void Assembler::IDIV( R r)  { emitr(X64_idiv, r); asm_output("idivl edx:eax, %s",RL(r)); }

	void Assembler::SHRL(R r)   { emitr(X64_shrl, r); asm_output("shrl %s, ecx", RL(r)); }
	void Assembler::SARL(R r)   { emitr(X64_sarl, r); asm_output("sarl %s, ecx", RL(r)); }
	void Assembler::SHLL(R r)   { emitr(X64_shll, r); asm_output("shll %s, ecx", RL(r)); }
	void Assembler::RORL(R r)   { emitr(X64_rorl, r); asm_output("rorl %s, ecx", RL(r)); }
	void Assembler::RCRL(R r)   { emitr(X64_rcrl, r); asm_output("rcrl %s, ecx", RL(r)); }
	void Assembler::SHRQ(R r)   { emitr(X64_shrq, r); asm_output("shrq %s, ecx", RQ(r)); }
	void Assembler::SARQ(R r)   { emitr(X64_sarq, r); asm_output("sarq %s, ecx", RQ(r)); }
	void Assembler::SHLQ(R r)   { emitr(X64_shlq, r); asm_output("shlq %s, ecx", RQ(r)); }
	void Assembler::RORQ(R r)   { emitr(X64_rorq, r); asm_output("rorq %s, ecx", RQ(r)); }
	void Assembler::RCRQ(R r)   { emitr(X64_rcrq, r); asm_output("rcrq %s, ecx", RQ(r)); }

	void Assembler::SHRLI( R r, I i)    { emit8(rexrb(X64_shrli | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("shrl %s, %d", RL(r), i); }
	void Assembler::SARLI( R r, I i)    { emit8(rexrb(X64_sarli | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("sarl %s, %d", RL(r), i); }
	void Assembler::SHLLI( R r, I i)    { emit8(rexrb(X64_shlli | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("shll %s, %d", RL(r), i); }
	void Assembler::RORLI( R r, I i)    { emit8(rexrb(X64_rorli | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("rorl %s, %d", RL(r), i); }
	void Assembler::RCRLI( R r, I i)    { emit8(rexrb(X64_rcrli | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("rcrl %s, %d", RL(r), i); }
	void Assembler::SHRQI(R r, I i)     { emit8(rexrb(X64_shrqi | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("shrq %s, %d", RQ(r), i); }
	void Assembler::SARQI(R r, I i)     { emit8(rexrb(X64_sarqi | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("sarq %s, %d", RQ(r), i); }
	void Assembler::SHLQI(R r, I i)     { emit8(rexrb(X64_shlqi | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("shlq %s, %d", RQ(r), i); }
	void Assembler::RORQI(R r, I i)     { emit8(rexrb(X64_rorqi | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("rorq %s, %d", RQ(r), i); }
	void Assembler::RCRQI(R r, I i)     { emit8(rexrb(X64_rcrqi | U64(REGNUM(r)&7)<<48, RZero, r), i); asm_output("rcrq %s, %d", RQ(r), i); }

	void Assembler::SETE( R r)  { emitr8(X64_sete, r); asm_output("sete %s", RB(r)); }
	void Assembler::SETL( R r)  { emitr8(X64_setl, r); asm_output("setl %s", RB(r)); }
	void Assembler::SETLE(R r)  { emitr8(X64_setle,r); asm_output("setle %s",RB(r)); }
	void Assembler::SETG( R r)  { emitr8(X64_setg, r); asm_output("setg %s", RB(r)); }
	void Assembler::SETGE(R r)  { emitr8(X64_setge,r); asm_output("setge %s",RB(r)); }
	void Assembler::SETB( R r)  { emitr8(X64_setb, r); asm_output("setb %s", RB(r)); }
	void Assembler::SETBE(R r)  { emitr8(X64_setbe,r); asm_output("setbe %s",RB(r)); }
	void Assembler::SETA( R r)  { emitr8(X64_seta, r); asm_output("seta %s", RB(r)); }
	void Assembler::SETAE(R r)  { emitr8(X64_setae,r); asm_output("setae %s",RB(r)); }
	void Assembler::SETO( R r)  { emitr8(X64_seto, r); asm_output("seto %s", RB(r)); }
	void Assembler::SETC( R r)  { emitr8(X64_setc, r); asm_output("setc %s", RB(r)); }
	void Assembler::SETZ( R r)  { emitr8(X64_setz, r); asm_output("setz %s", RB(r)); };
	void Assembler::SETS( R r)  { emitr8(X64_sets, r); asm_output("sets %s", RB(r)); };
	void Assembler::SETNO( R r) { emitr8(X64_seto ^ X64_setneg, r); asm_output("setno %s", RB(r)); };
	void Assembler::SETNC( R r) { emitr8(X64_setc ^ X64_setneg, r); asm_output("setnc %s", RB(r)); };
	void Assembler::SETNZ( R r) { emitr8(X64_setz ^ X64_setneg, r); asm_output("setnz %s", RB(r)); };
	void Assembler::SETNS( R r) { emitr8(X64_sets ^ X64_setneg, r); asm_output("setns %s", RB(r)); };

	void Assembler::ADDRR(R l, R r)     { emitrr(X64_addlrr,l,r); asm_output("addl %s, %s", RL(l),RL(r)); }
	void Assembler::ADCRR(R l, R r)     { emitrr(X64_adclrr,l,r); asm_output("adcl %s, %s", RL(l),RL(r)); }
	void Assembler::SUBRR(R l, R r)     { emitrr(X64_sublrr,l,r); asm_output("subl %s, %s", RL(l),RL(r)); }
	void Assembler::SBBRR(R l, R r)     { emitrr(X64_sbblrr,l,r); asm_output("sbbl %s, %s", RL(l),RL(r)); }
	void Assembler::ANDRR(R l, R r)     { emitrr(X64_andrr,l,r); asm_output("andl %s, %s", RL(l),RL(r)); }
	void Assembler::ORLRR(R l, R r)     { emitrr(X64_orlrr,l,r); asm_output("orl %s, %s",  RL(l),RL(r)); }
	void Assembler::XORRR(R l, R r)     { emitrr(X64_xorrr,l,r); asm_output("xorl %s, %s", RL(l),RL(r)); }
	void Assembler::IMUL( R l, R r)     { emitrr(X64_imul, l,r); asm_output("imul %s, %s",RL(l),RL(r)); }
	void Assembler::IMULQ( R l, R r)    { emitrr(X64_imulq, l,r); asm_output("imulq %s, %s",RL(l),RL(r)); }
	void Assembler::SMULL( R r)         { emitr(X64_smull, r); asm_output("smull %s:%s, %s",RL(RAX),RL(RDX),RL(r)); }
	void Assembler::UMULL( R r)         { emitr(X64_mull, r); asm_output("mull %s:%s, %s",RL(RAX),RL(RDX),RL(r)); }
	void Assembler::SMULLQ( R r)        { emitr(X64_smullq, r); asm_output("smull %s:%s, %s",RQ(RAX),RQ(RDX),RQ(r)); }
	void Assembler::UMULLQ( R r)        { emitr(X64_mullq, r); asm_output("mull %s:%s, %s",RQ(RAX),RQ(RDX),RQ(r)); }

	void Assembler::CMPLR(R l, R r)     { emitrr(X64_cmplr,l,r); asm_output("cmpl %s, %s", RL(l),RL(r)); }
	void Assembler::MOVLR(R l, R r)     { emitrr(X64_movlr,l,r); asm_output("movl %s, %s", RL(l),RL(r)); }

	void Assembler::ADDQRR( R l, R r)   { emitrr(X64_addqrr, l,r); asm_output("addq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ADCQRR( R l, R r)   { emitrr(X64_adcqrr, l,r); asm_output("adcq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::SUBQRR( R l, R r)   { emitrr(X64_subqrr, l,r); asm_output("subq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::SBBQRR( R l, R r)   { emitrr(X64_sbbqrr, l,r); asm_output("sbbq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ANDQRR( R l, R r)   { emitrr(X64_andqrr, l,r); asm_output("andq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ORQRR(  R l, R r)   { emitrr(X64_orqrr,  l,r); asm_output("orq %s, %s",   RQ(l),RQ(r)); }
	void Assembler::XORQRR( R l, R r)   { emitrr(X64_xorqrr, l,r); asm_output("xorq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPQR(  R l, R r)   { emitrr(X64_cmpqr,  l,r); asm_output("cmpq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::MOVQR(  R l, R r)   { emitrr(X64_movqr,  l,r); asm_output("movq %s, %s",  RQ(l),RQ(r)); }
	void Assembler::MOVAPSR(R l, R r)   { emitrr(X64_movapsr,l,r); asm_output("movaps %s, %s",RQ(l),RQ(r)); }
	void Assembler::UNPCKLPS(R l, R r)  { emitrr(X64_unpcklps,l,r);asm_output("unpcklps %s, %s",RQ(l),RQ(r));}

	void Assembler::CMOVE(  R l, R r)   { emitrr(X64_cmove,  l,r); asm_output("cmovle %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNO( R l, R r)   { emitrr(X64_cmovno, l,r); asm_output("cmovlno %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNE( R l, R r)   { emitrr(X64_cmovne, l,r); asm_output("cmovlne %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNL( R l, R r)   { emitrr(X64_cmovnl, l,r); asm_output("cmovlnl %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNLE(R l, R r)   { emitrr(X64_cmovnle,l,r); asm_output("cmovlnle %s, %s", RL(l),RL(r)); }
	void Assembler::CMOVNG( R l, R r)   { emitrr(X64_cmovng, l,r); asm_output("cmovlng %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNGE(R l, R r)   { emitrr(X64_cmovnge,l,r); asm_output("cmovlnge %s, %s", RL(l),RL(r)); }
	void Assembler::CMOVNB( R l, R r)   { emitrr(X64_cmovnb, l,r); asm_output("cmovlnb %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNBE(R l, R r)   { emitrr(X64_cmovnbe,l,r); asm_output("cmovlnbe %s, %s", RL(l),RL(r)); }
	void Assembler::CMOVNA( R l, R r)   { emitrr(X64_cmovna, l,r); asm_output("cmovlna %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVNAE(R l, R r)   { emitrr(X64_cmovnae,l,r); asm_output("cmovlnae %s, %s", RL(l),RL(r)); }

	void Assembler::CMOVNP(R l, R r)   { emitrr(X64_cmovnp,l,r); asm_output("cmovlnp %s, %s", RL(l),RL(r)); }

	void Assembler::CMOVQE(  R l, R r)  { emitrr(X64_cmovqe,  l,r); asm_output("cmovqe %s, %s",  RL(l),RL(r)); }
	void Assembler::CMOVQNO( R l, R r)  { emitrr(X64_cmovqno, l,r); asm_output("cmovqno %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNE( R l, R r)  { emitrr(X64_cmovqne, l,r); asm_output("cmovqne %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNL( R l, R r)  { emitrr(X64_cmovqnl, l,r); asm_output("cmovqnl %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNLE(R l, R r)  { emitrr(X64_cmovqnle,l,r); asm_output("cmovqnle %s, %s", RQ(l),RQ(r)); }
	void Assembler::CMOVQNG( R l, R r)  { emitrr(X64_cmovqng, l,r); asm_output("cmovqng %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNGE(R l, R r)  { emitrr(X64_cmovqnge,l,r); asm_output("cmovqnge %s, %s", RQ(l),RQ(r)); }
	void Assembler::CMOVQNB( R l, R r)  { emitrr(X64_cmovqnb, l,r); asm_output("cmovqnb %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNBE(R l, R r)  { emitrr(X64_cmovqnbe,l,r); asm_output("cmovqnbe %s, %s", RQ(l),RQ(r)); }
	void Assembler::CMOVQNA( R l, R r)  { emitrr(X64_cmovqna, l,r); asm_output("cmovqna %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMOVQNAE(R l, R r)  { emitrr(X64_cmovqnae,l,r); asm_output("cmovqnae %s, %s", RQ(l),RQ(r)); }

	void Assembler::CMOVQNP(R l, R r)  { emitrr(X64_cmovqnp,l,r); asm_output("cmovqnp %s, %s", RQ(l),RQ(r)); }

	void Assembler::MOVSXDR(R l, R r)   { emitrr(X64_movsxdr,l,r); asm_output("movsxd %s, %s",RQ(l),RL(r)); }

	void Assembler::MOVZX8(R l, R r)    { emitrr8(X64_movzx8,l,r); asm_output("movzx %s, %s",RQ(l),RB(r)); }

// XORPS is a 4x32f vector operation, we use it instead of the more obvious
// XORPD because it's one byte shorter.  This is ok because it's only used for
// zeroing an XMM register;  hence the single argument.
// Also note that (unlike most SSE2 instructions) XORPS does not have a prefix, thus emitrr() should be used.
	void Assembler::XORPS(   R l, R r)  { emitrr(X64_xorps,    l,r); asm_output("xorps %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ANDPS(   R l, R r)  { emitrr(X64_andps,    l,r); asm_output("andps %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ANDNPS(  R l, R r)  { emitrr(X64_andnps,   l,r); asm_output("andnps %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ANDPD(   R l, R r)  { emitprr(X64_andpd,    l,r); asm_output("andpd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ANDNPD(  R l, R r)  { emitprr(X64_andnpd,   l,r); asm_output("andnpd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ORPS(    R l, R r)  { emitrr(X64_orps,     l,r); asm_output("orps %s, %s",   RQ(l),RQ(r)); }
	void Assembler::ORPD(    R l, R r)  { emitprr(X64_orpd,     l,r); asm_output("orpd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::DIVSD(   R l, R r)  { emitprr(X64_divsd,   l,r); asm_output("divsd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::MULSD(   R l, R r)  { emitprr(X64_mulsd,   l,r); asm_output("mulsd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::SQRTSD(  R l, R r)  { emitprr(X64_sqrtsd,  l,r); asm_output("sqrtsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ADDSD(   R l, R r)  { emitprr(X64_addsd,   l,r); asm_output("addsd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::SUBSD(   R l, R r)  { emitprr(X64_subsd,   l,r); asm_output("subsd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::DIVSS(   R l, R r)  { emitprr(X64_divss,   l,r); asm_output("divss %s, %s",   RQ(l),RQ(r)); }
	void Assembler::MULSS(   R l, R r)  { emitprr(X64_mulss,   l,r); asm_output("mulss %s, %s",   RQ(l),RQ(r)); }
	void Assembler::SQRTSS(  R l, R r)  { emitprr(X64_sqrtss,  l,r); asm_output("sqrtss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ADDSS(   R l, R r)  { emitprr(X64_addss,   l,r); asm_output("addss %s, %s",   RQ(l),RQ(r)); }
	void Assembler::SUBSS(   R l, R r)  { emitprr(X64_subss,   l,r); asm_output("subss %s, %s",   RQ(l),RQ(r)); }
	void Assembler::DIVPS(   R l, R r)  { emitrr(X64_divps,   l,r); asm_output("divps %s, %s",  RQ(l),RQ(r)); }
	void Assembler::MULPS(   R l, R r)  { emitrr(X64_mulps,   l,r); asm_output("mulps %s, %s",  RQ(l),RQ(r)); }
	void Assembler::ADDPS(   R l, R r)  { emitrr(X64_addps,   l,r); asm_output("addps %s, %s",  RQ(l),RQ(r)); }


	void Assembler::CVTSQ2SD(R l, R r)  { emitprr(X64_cvtsq2sd,l,r); asm_output("cvtsq2sd %s, %s",RQ(l),RQ(r)); }
	void Assembler::CVTSQ2SS(R l, R r)  { emitprr(X64_cvtsq2ss,l,r); asm_output("cvtsq2ss %s, %s",RQ(l),RQ(r)); }
	void Assembler::CVTSI2SD(R l, R r)  { emitprr(X64_cvtsi2sd,l,r); asm_output("cvtsi2sd %s, %s",RQ(l),RL(r)); }
	void Assembler::CVTSI2SS(R l, R r)  { emitprr(X64_cvtsi2ss,l,r); asm_output("cvtsi2ss %s, %s",RQ(l),RL(r)); }
	void Assembler::CVTSS2SD(R l, R r)  { emitprr(X64_cvtss2sd,l,r); asm_output("cvtss2sd %s, %s",RQ(l),RL(r)); }
	void Assembler::CVTSD2SS(R l, R r)  { emitprr(X64_cvtsd2ss,l,r); asm_output("cvtsd2ss %s, %s",RQ(l),RQ(r)); }

	void Assembler::CVTSD2SI(R l, R r)  { emitprr(X64_cvtsd2si,l,r); asm_output("cvtsd2si %s, %s",RL(l),RQ(r)); }
	void Assembler::CVTTSD2SI(R l, R r) { emitprr(X64_cvttsd2si,l,r);asm_output("cvttsd2si %s, %s",RL(l),RQ(r));}
	void Assembler::CVTSS2SI(R l, R r)  { emitprr(X64_cvtss2si,l,r);asm_output("cvtss2si %s, %s",RL(l),RQ(r));}
	void Assembler::CVTTSS2SI(R l, R r) { emitprr(X64_cvttss2si,l,r);asm_output("cvttss2si %s, %s",RL(l),RQ(r));}

	void Assembler::CVTSD2SQ(R l, R r)  { emitprr(X64_cvtsd2sq,l,r); asm_output("cvtsd2si %s, %s",RL(l),RQ(r)); }
	void Assembler::CVTTSD2SQ(R l, R r) { emitprr(X64_cvttsd2sq,l,r);asm_output("cvttsd2si %s, %s",RL(l),RQ(r));}
	void Assembler::CVTSS2SQ(R l, R r)  { emitprr(X64_cvtss2sq,l,r);asm_output("cvtss2si %s, %s",RL(l),RQ(r));}
	void Assembler::CVTTSS2SQ(R l, R r) { emitprr(X64_cvttss2sq,l,r);asm_output("cvttss2si %s, %s",RL(l),RQ(r));}
	
	void Assembler::CVTPS2DQ(R l, R r)  { emitprr(X64_cvtps2dq,l,r);asm_output("cvtps2dq %s, %s",RL(l),RQ(r));}
	void Assembler::CVTTPS2DQ(R l, R r) { emitprr(X64_cvttps2dq,l,r);asm_output("cvttps2dq %s, %s",RL(l),RQ(r));}
	void Assembler::CVTPD2DQ(R l, R r)  { emitprr(X64_cvtpd2dq,l,r);asm_output("cvtpd2dq %s, %s",RL(l),RQ(r));}
	void Assembler::CVTTPD2DQ(R l, R r) { emitprr(X64_cvttpd2dq,l,r);asm_output("cvttpd2dq %s, %s",RL(l),RQ(r));}

	void Assembler::CVTDQ2PS(R l, R r)  { emitrr(X64_cvtdq2ps,l,r);asm_output("cvtdq2ps %s, %s",RQ(l),RQ(r));}
		
	void Assembler::UCOMISS( R l, R r)  { emitrr(X64_ucomiss, l,r);  asm_output("ucomiss %s, %s", RQ(l),RQ(r)); }
	void Assembler::UCOMISD( R l, R r)  { emitprr(X64_ucomisd, l,r); asm_output("ucomisd %s, %s", RQ(l),RQ(r)); }
	void Assembler::MOVQRX(  R l, R r)  { emitprr(X64_movqrx,  r,l); asm_output("movq %s, %s",   RQ(l),RQ(r)); } // Nb: r and l are deliberately reversed within the emitprr()  call.
	void Assembler::MOVQXR(  R l, R r)  { emitprr(X64_movqxr,  l,r); asm_output("movq %s, %s",   RQ(l),RQ(r)); }
	void Assembler::MOVDXR(  R l, R r)  { emitprr(X64_movdxr,  l,r); asm_output("movd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::MOVDRX(  R l, R r)  { emitprr(X64_movdrx,  l,r); asm_output("movd %s, %s",   RQ(l),RQ(r)); }
	void Assembler::MOVLHPS( R l, R r)  { emitrr(X64_movlhps, l,r);  asm_output("movlhps %s, %s", RQ(l),RQ(r)); }
	void Assembler::PMOVMSKB(R l, R r)  { emitprr(X64_pmovmskb,l,r); asm_output("pmovmskb %s, %s",RQ(l),RQ(r)); }

	void Assembler::CMPEQSS(    R l, R r)  { emitprr_imm8(X64_cmpss,l,r,0); asm_output("cmpeqss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPLTSS(    R l, R r)  { emitprr_imm8(X64_cmpss,l,r,1); asm_output("cmpltss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPLESS(    R l, R r)  { emitprr_imm8(X64_cmpss,l,r,2); asm_output("cmpless %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPUNORDSS( R l, R r)  { emitprr_imm8(X64_cmpss,l,r,3); asm_output("cmpunordss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNEQSS(   R l, R r)  { emitprr_imm8(X64_cmpss,l,r,4); asm_output("cmpneqss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNLTSS(   R l, R r)  { emitprr_imm8(X64_cmpss,l,r,5); asm_output("cmpnltss %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNLESS(   R l, R r)  { emitprr_imm8(X64_cmpss,l,r,6); asm_output("cmpnless %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPORDSS(   R l, R r)  { emitprr_imm8(X64_cmpss,l,r,7); asm_output("cmpordess %s, %s",  RQ(l),RQ(r)); }

	void Assembler::CMPEQSD(    R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,0); asm_output("cmpeqsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPLTSD(    R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,1); asm_output("cmpltsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPLESD(    R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,2); asm_output("cmplesd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPUNORDSD( R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,3); asm_output("cmpunordsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNEQSD(   R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,4); asm_output("cmpneqsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNLTSD(   R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,5); asm_output("cmpnltsd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPNLESD(   R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,6); asm_output("cmpnlesd %s, %s",  RQ(l),RQ(r)); }
	void Assembler::CMPPRDSD(   R l, R r)  { emitprr_imm8(X64_cmpsd,l,r,7); asm_output("cmpordesd %s, %s",  RQ(l),RQ(r)); }


	inline uint8_t PSHUFD_MASK(int x, int y, int z, int w) {
		NanoAssert(x>=0 && x<=3);
		NanoAssert(y>=0 && y<=3);
		NanoAssert(z>=0 && z<=3);
		NanoAssert(w>=0 && w<=3);
		return (uint8_t) (x | (y<<2) | (z<<4) | (w<<6));
	}
	void Assembler::PSHUFD(R l, R r, I m) {
		NanoAssert(isU8(m));
		uint8_t mode = uint8_t(m);
		emitprr_imm8(X64_pshufd, l, r, mode);
		asm_output("pshufd  %s, %s, %x", RQ(l), RQ(r), m);
	}
	void Assembler::SHUFPD(R l, R r, I m)  {
		NanoAssert(m>=0 && m<256);
		NanoAssert(isU8(m));
		uint8_t mode = (uint8_t) m;
		emitprr_imm8(X64_shufpd,l,r,mode);
		asm_output("shufpd  %s, %s, %x", RQ(l), RQ(r), m);
	}

	// MOVI must not affect condition codes!
	void Assembler::MOVI(  R r, I32 i32)    { emitr_imm(X64_movi,  r,i32); asm_output("movl %s, %d",RL(r),i32); }
	void Assembler::ADDLRI(R r, I32 i32)    { emitr_imm(X64_addlri,r,i32); asm_output("addl %s, %d",RL(r),i32); }
	void Assembler::ADCLRI(R r, I32 i32)    { emitr_imm(X64_adclri,r,i32); asm_output("adcl %s, %d",RL(r),i32); }
	void Assembler::SUBLRI(R r, I32 i32)    { emitr_imm(X64_sublri,r,i32); asm_output("subl %s, %d",RL(r),i32); }
	void Assembler::SBBLRI(R r, I32 i32)    { emitr_imm(X64_sbblri,r,i32); asm_output("sbbl %s, %d",RL(r),i32); }
	void Assembler::ANDLRI(R r, I32 i32)    { emitr_imm(X64_andlri,r,i32); asm_output("andl %s, %d",RL(r),i32); }
	void Assembler::ORLRI( R r, I32 i32)    { emitr_imm(X64_orlri, r,i32); asm_output("orl %s, %d", RL(r),i32); }
	void Assembler::XORLRI(R r, I32 i32)    { emitr_imm(X64_xorlri,r,i32); asm_output("xorl %s, %d",RL(r),i32); }
	void Assembler::CMPLRI(R r, I32 i32)    { emitr_imm(X64_cmplri,r,i32); asm_output("cmpl %s, %d",RL(r),i32); }

	void Assembler::ADDQRI( R r, I32 i32)   { emitr_imm(X64_addqri, r,i32); asm_output("addq %s, %d", RQ(r),i32); }
	void Assembler::ADCQRI( R r, I32 i32)   { emitr_imm(X64_adcqri, r,i32); asm_output("adcq %s, %d", RQ(r),i32); }
	void Assembler::SUBQRI( R r, I32 i32)   { emitr_imm(X64_subqri, r,i32); asm_output("subq %s, %d", RQ(r),i32); }
	void Assembler::SBBQRI( R r, I32 i32)   { emitr_imm(X64_sbbqri, r,i32); asm_output("sbbq %s, %d", RQ(r),i32); }
	void Assembler::ANDQRI( R r, I32 i32)   { emitr_imm(X64_andqri, r,i32); asm_output("andq %s, %d", RQ(r),i32); }
	void Assembler::ORQRI(  R r, I32 i32)   { emitr_imm(X64_orqri,  r,i32); asm_output("orq %s, %d", RQ(r),i32); }
	void Assembler::XORQRI( R r, I32 i32)   { emitr_imm(X64_xorqri, r,i32); asm_output("xorq %s, %d", RQ(r),i32); }
	void Assembler::CMPQRI( R r, I32 i32)   { emitr_imm(X64_cmpqri, r,i32); asm_output("cmpq %s, %d", RQ(r),i32); }
	void Assembler::MOVQI32(R r, I32 i32)   { emitr_imm(X64_movqi32,r,i32); asm_output("movqi32 %s, %d",RQ(r),i32); }

	void Assembler::ADDLR8(R r, I32 i8)     { emitr_imm8(X64_addlr8,r,i8); asm_output("addl %s, %d", RL(r),i8); }
	void Assembler::ADCLR8(R r, I32 i8)     { emitr_imm8(X64_adclr8,r,i8); asm_output("adcl %s, %d", RL(r),i8); }
	void Assembler::SUBLR8(R r, I32 i8)     { emitr_imm8(X64_sublr8,r,i8); asm_output("subl %s, %d", RL(r),i8); }
	void Assembler::SBBLR8(R r, I32 i8)     { emitr_imm8(X64_sbblr8,r,i8); asm_output("sbbl %s, %d", RL(r),i8); }
	void Assembler::ANDLR8(R r, I32 i8)     { emitr_imm8(X64_andlr8,r,i8); asm_output("andl %s, %d", RL(r),i8); }
	void Assembler::ORLR8( R r, I32 i8)     { emitr_imm8(X64_orlr8, r,i8); asm_output("orl %s, %d",  RL(r),i8); }
	void Assembler::XORLR8(R r, I32 i8)     { emitr_imm8(X64_xorlr8,r,i8); asm_output("xorl %s, %d", RL(r),i8); }
	void Assembler::CMPLR8(R r, I32 i8)     { emitr_imm8(X64_cmplr8,r,i8); asm_output("cmpl %s, %d", RL(r),i8); }

	void Assembler::ADDQR8(R r, I32 i8)     { emitr_imm8(X64_addqr8,r,i8); asm_output("addq %s, %d",RQ(r),i8); }
	void Assembler::ADCQR8(R r, I32 i8)     { emitr_imm8(X64_adcqr8,r,i8); asm_output("adcq %s, %d",RQ(r),i8); }
	void Assembler::SUBQR8(R r, I32 i8)     { emitr_imm8(X64_subqr8,r,i8); asm_output("subq %s, %d",RQ(r),i8); }
	void Assembler::SBBQR8(R r, I32 i8)     { emitr_imm8(X64_sbbqr8,r,i8); asm_output("sbbq %s, %d",RQ(r),i8); }
	void Assembler::ANDQR8(R r, I32 i8)     { emitr_imm8(X64_andqr8,r,i8); asm_output("andq %s, %d",RQ(r),i8); }
	void Assembler::ORQR8( R r, I32 i8)     { emitr_imm8(X64_orqr8, r,i8); asm_output("orq %s, %d", RQ(r),i8); }
	void Assembler::XORQR8(R r, I32 i8)     { emitr_imm8(X64_xorqr8,r,i8); asm_output("xorq %s, %d",RQ(r),i8); }
	void Assembler::CMPQR8(R r, I32 i8)     { emitr_imm8(X64_cmpqr8,r,i8); asm_output("cmpq %s, %d",RQ(r),i8); }

	void Assembler::IMULI(R l, R r, I32 i32)        { emitrr_imm(X64_imuli,l,r,i32); asm_output("imuli %s, %s, %d",RL(l),RL(r),i32); }
	void Assembler::IMULIQ(R    l, R r, I32 i32)    { emitrr_imm(X64_imuliq,l,r,i32); asm_output("imuli %s, %s, %d",RL(l),RL(r),i32); }

	void Assembler::MOVQI(R r, U64 u64)         { emitr_imm64(X64_movqi,r,u64); asm_output("movq %s, %p",RQ(r),(void*)u64); }

	void Assembler::LEARIP(R r, I32 d)          { emitrm(X64_learip,r,d,RZero); asm_output("lea %s, %d(rip)",RQ(r),d); }

	void Assembler::LEALRM(R r, I d, R b)       { emitrm(X64_lealrm,r,d,b); asm_output("leal %s, %d(%s)",RL(r),d,RL(b)); }
	void Assembler::LEAQRM(R r, I d, R b)       { emitrm(X64_leaqrm,r,d,b); asm_output("leaq %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVLRM(R r, I d, R b)       { emitrm(X64_movlrm,r,d,b); asm_output("movl %s, %d(%s)",RL(r),d,RQ(b)); }
	void Assembler::MOVQRM(R r, I d, R b)       { emitrm(X64_movqrm,r,d,b); asm_output("movq %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVBMR(R r, I d, R b)       { emitrm8(X64_movbmr,r,d,b); asm_output("movb %d(%s), %s",d,RQ(b),RB(r)); }
	void Assembler::MOVSMR(R r, I d, R b)       { emitprm(X64_movsmr,r,d,b); asm_output("movs %d(%s), %s",d,RQ(b),RS(r)); }
	void Assembler::MOVLMR(R r, I d, R b)       { emitrm(X64_movlmr,r,d,b); asm_output("movl %d(%s), %s",d,RQ(b),RL(r)); }
	void Assembler::MOVQMR(R r, I d, R b)       { emitrm(X64_movqmr,r,d,b); asm_output("movq %d(%s), %s",d,RQ(b),RQ(r)); }

	void Assembler::MOVZX8M( R r, I d, R b)     { emitrm_wide(X64_movzx8m, r,d,b); asm_output("movzxb %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVZX16M(R r, I d, R b)     { emitrm_wide(X64_movzx16m,r,d,b); asm_output("movzxs %s, %d(%s)",RQ(r),d,RQ(b)); }

	void Assembler::MOVSX8M( R r, I d, R b)     { emitrm_wide(X64_movsx8m, r,d,b); asm_output("movsxb %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVSX16M(R r, I d, R b)     { emitrm_wide(X64_movsx16m,r,d,b); asm_output("movsxs %s, %d(%s)",RQ(r),d,RQ(b)); }

	void Assembler::MOVSDRM(R r, I d, R b)      { emitprm(X64_movsdrm,r,d,b); asm_output("movsd %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVSDMR(R r, I d, R b)      { emitprm(X64_movsdmr,r,d,b); asm_output("movsd %d(%s), %s",d,RQ(b),RQ(r)); }
	void Assembler::MOVSSRM(R r, I d, R b)      { emitprm(X64_movssrm,r,d,b); asm_output("movss %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVSSMR(R r, I d, R b)      { emitprm(X64_movssmr,r,d,b); asm_output("movss %d(%s), %s",d,RQ(b),RQ(r)); }
	void Assembler::MOVUPSRM(R r, I d, R b)     { emitrm_wide(X64_movupsrm,r,d,b); asm_output("movups %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVUPSMR(R r, I d, R b)     { emitrm_wide(X64_movupsmr,r,d,b); asm_output("movups %d(%s), %s",d,RQ(b),RQ(r)); }
	void Assembler::MOVUPSRMRIP(R r, I d)       { emitrm_wide(X64_movupsrip,r,d,RZero); asm_output("movups %s, %d(rip)",RQ(r),d); }
	void Assembler::MOVAPSRM(R r, I d, R b)     { emitrm_wide(X64_movapsrm,r,d,b); asm_output("movaps %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVAPSRMRIP(R r, I d)       { emitrm_wide(X64_movapsrip,r,d,RZero); asm_output("movaps %s, %d(rip)",RQ(r),d); }
	void Assembler::MOVSSSPR(R r, I d)          {
												  uint64_t op = emit_disp32_sib(X64_movssspr,d);
												  emit( op | U64((REGNUM(r)&7)<<3) << 48 | U64((REGNUM(r)&8)>>1) << 24);
												  asm_output("movss %s, %d(RSP)",RQ(r),d);
												}
	void Assembler::MOVSDSPR(R r, I d)          {
												  uint64_t op = emit_disp32_sib(X64_movsdspr,d);
												  emit( op | U64((REGNUM(r)&7)<<3) << 48 | U64((REGNUM(r)&8)>>1) << 24);
												  asm_output("movsd %s, %d(RSP)",RQ(r),d);
												}
	void Assembler::MOVUPSSPR(R r, I d)         {
												  uint64_t op = emit_disp32_sib(X64_movupspr,d);
												  emit( op | U64((REGNUM(r)&7)<<3) << 48 | U64((REGNUM(r)&8)>>1) << 24);
												  asm_output("movups %s, %d(RSP)",RQ(r),d);
												}


	void Assembler::MOVSDRM32(R r, I d, R   b)      { emitprm32(X64_movsdrm,r,d,b);         asm_output("movsd %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVQRM32(R r,   I d, R b)       { emitrm32(X64_movqrm,r,d,b);           asm_output("movq %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVSSRM32(R r, I d, R   b)      { emitprm32(X64_movssrm,r,d,b);         asm_output("movss %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVUPSRM32(R r, I   d, R b)     { emitrm_wide32(X64_movupsrm,r,d,b);    asm_output("movups %s, %d(%s)",RQ(r),d,RQ(b)); }
	void Assembler::MOVLRM32(R r,   I d, R b)       { emitrm32(X64_movlrm,r,d,b);           asm_output("movl %s, %d(%s)",RL(r),d,RQ(b)); }


	void Assembler::JMP8( S n, const NIns* t)       { emit_target8(n, X64_jmp8,t); asm_output("jmp %p", t); }

	void Assembler::JMP32(S n, const NIns* t)       { emit_target32(n,X64_jmp, t); asm_output("jmp %p", t); }
	void Assembler::JMP64(S n)                      { emit_targetIpRel(n, X64_jmpi, 0); asm_output("jmp [ip]"); }

	void Assembler::JMPXB(R indexreg, R tablereg)   { emitxb(X64_jmpxb, indexreg, tablereg); asm_output("jmp [%s*8 + %s]", RQ(indexreg), RQ(tablereg)); }

	void Assembler::JO( S n, const NIns* t)         { emit_target32(n,X64_jo,  t); asm_output("jo %p", t); }
	void Assembler::JE( S n, const NIns* t)         { emit_target32(n,X64_je,  t); asm_output("je %p", t); }
	void Assembler::JL( S n, const NIns* t)         { emit_target32(n,X64_jl,  t); asm_output("jl %p", t); }
	void Assembler::JLE(S n, const NIns* t)         { emit_target32(n,X64_jle, t); asm_output("jle %p",t); }
	void Assembler::JG( S n, const NIns* t)         { emit_target32(n,X64_jg,  t); asm_output("jg %p", t); }
	void Assembler::JGE(S n, const NIns* t)         { emit_target32(n,X64_jge, t); asm_output("jge %p",t); }
	void Assembler::JB( S n, const NIns* t)         { emit_target32(n,X64_jb,  t); asm_output("jb %p", t); }
	void Assembler::JBE(S n, const NIns* t)         { emit_target32(n,X64_jbe, t); asm_output("jbe %p",t); }
	void Assembler::JA( S n, const NIns* t)         { emit_target32(n,X64_ja,  t); asm_output("ja %p", t); }
	void Assembler::JAE(S n, const NIns* t)         { emit_target32(n,X64_jae, t); asm_output("jae %p",t); }
	void Assembler::JP( S n, const NIns* t)         { emit_target32(n,X64_jp,  t); asm_output("jp  %p",t); }
	void Assembler::JC( S n, const NIns* t)         { emit_target8(n,X64_jb,  t); asm_output("jc  %p",t); }
	void Assembler::JZ( S n, const NIns* t)         { emit_target8(n,X64_je,  t); asm_output("jz  %p",t); }
	void Assembler::JS( S n, const NIns* t)         { emit_target8(n,X64_js,  t); asm_output("js  %p",t); }

	void Assembler::JNO( S n, const NIns* t)        { emit_target32(n,X64_jo  ^ X64_jneg, t); asm_output("jno %p", t); }
	void Assembler::JNE( S n, const NIns* t)        { emit_target32(n,X64_je  ^ X64_jneg, t); asm_output("jne %p", t); }
	void Assembler::JNL( S n, const NIns* t)        { emit_target32(n,X64_jl  ^ X64_jneg, t); asm_output("jnl %p", t); }
	void Assembler::JNLE(S n, const NIns* t)        { emit_target32(n,X64_jle ^ X64_jneg, t); asm_output("jnle %p",t); }
	void Assembler::JNG( S n, const NIns* t)        { emit_target32(n,X64_jg  ^ X64_jneg, t); asm_output("jng %p", t); }
	void Assembler::JNGE(S n, const NIns* t)        { emit_target32(n,X64_jge ^ X64_jneg, t); asm_output("jnge %p",t); }
	void Assembler::JNB( S n, const NIns* t)        { emit_target32(n,X64_jb  ^ X64_jneg, t); asm_output("jnb %p", t); }
	void Assembler::JNBE(S n, const NIns* t)        { emit_target32(n,X64_jbe ^ X64_jneg, t); asm_output("jnbe %p",t); }
	void Assembler::JNA( S n, const NIns* t)        { emit_target32(n,X64_ja  ^ X64_jneg, t); asm_output("jna %p", t); }
	void Assembler::JNAE(S n, const NIns* t)        { emit_target32(n,X64_jae ^ X64_jneg, t); asm_output("jnae %p",t); }
	void Assembler::JNC( S n, const NIns* t)        { emit_target8(n,X64_jb ^ X64_jneg,  t); asm_output("jnc %p",t); }
	void Assembler::JNZ( S n, const NIns* t)        { emit_target8(n,X64_je ^ X64_jneg,  t); asm_output("jnz %p",t); }
	void Assembler::JNS( S n, const NIns* t)        { emit_target8(n,X64_js ^ X64_jneg,  t); asm_output("jns %p",t); }

	void Assembler::JO8( S n, const NIns* t)        { emit_target8(n,X64_jo8,  t); asm_output("jo %p", t); }
	void Assembler::JE8( S n, const NIns* t)        { emit_target8(n,X64_je8,  t); asm_output("je %p", t); }
	void Assembler::JL8( S n, const NIns* t)        { emit_target8(n,X64_jl8,  t); asm_output("jl %p", t); }
	void Assembler::JLE8(S n, const NIns* t)        { emit_target8(n,X64_jle8, t); asm_output("jle %p",t); }
	void Assembler::JG8( S n, const NIns* t)        { emit_target8(n,X64_jg8,  t); asm_output("jg %p", t); }
	void Assembler::JGE8(S n, const NIns* t)        { emit_target8(n,X64_jge8, t); asm_output("jge %p",t); }
	void Assembler::JB8( S n, const NIns* t)        { emit_target8(n,X64_jb8,  t); asm_output("jb %p", t); }
	void Assembler::JBE8(S n, const NIns* t)        { emit_target8(n,X64_jbe8, t); asm_output("jbe %p",t); }
	void Assembler::JA8( S n, const NIns* t)        { emit_target8(n,X64_ja8,  t); asm_output("ja %p", t); }
	void Assembler::JAE8(S n, const NIns* t)        { emit_target8(n,X64_jae8, t); asm_output("jae %p",t); }
	void Assembler::JP8( S n, const NIns* t)        { emit_target8(n,X64_jp8,  t); asm_output("jp %p",t); }
	void Assembler::JC8( S n, const NIns* t)        { emit_target8(n,X64_jb8,  t); asm_output("jc %p",t); }
	void Assembler::JZ8( S n, const NIns* t)        { emit_target8(n,X64_je8,  t); asm_output("jz %p",t); }
	void Assembler::JS8( S n, const NIns* t)        { emit_target8(n,X64_js8,  t); asm_output("js %p",t); }

	void Assembler::JNO8( S n, const NIns* t)       { emit_target8(n,X64_jo8  ^ X64_jneg8, t); asm_output("jno %p", t); }
	void Assembler::JNE8( S n, const NIns* t)       { emit_target8(n,X64_je8  ^ X64_jneg8, t); asm_output("jne %p", t); }
	void Assembler::JNL8( S n, const NIns* t)       { emit_target8(n,X64_jl8  ^ X64_jneg8, t); asm_output("jnl %p", t); }
	void Assembler::JNLE8(S n, const NIns* t)       { emit_target8(n,X64_jle8 ^ X64_jneg8, t); asm_output("jnle %p",t); }
	void Assembler::JNG8( S n, const NIns* t)       { emit_target8(n,X64_jg8  ^ X64_jneg8, t); asm_output("jng %p", t); }
	void Assembler::JNGE8(S n, const NIns* t)       { emit_target8(n,X64_jge8 ^ X64_jneg8, t); asm_output("jnge %p",t); }
	void Assembler::JNB8( S n, const NIns* t)       { emit_target8(n,X64_jb8  ^ X64_jneg8, t); asm_output("jnb %p", t); }
	void Assembler::JNBE8(S n, const NIns* t)       { emit_target8(n,X64_jbe8 ^ X64_jneg8, t); asm_output("jnbe %p",t); }
	void Assembler::JNA8( S n, const NIns* t)       { emit_target8(n,X64_ja8  ^ X64_jneg8, t); asm_output("jna %p", t); }
	void Assembler::JNAE8(S n, const NIns* t)       { emit_target8(n,X64_jae8 ^ X64_jneg8, t); asm_output("jnae %p",t); }
	void Assembler::JNC8( S n, const NIns* t)       { emit_target8(n,X64_jb8 ^ X64_jneg8,  t); asm_output("jnc %p",t); }
	void Assembler::JNZ8( S n, const NIns* t)       { emit_target8(n,X64_je8 ^ X64_jneg8,  t); asm_output("jnz %p",t); }
	void Assembler::JNS8( S n, const NIns* t)       { emit_target8(n,X64_js8 ^ X64_jneg8,  t); asm_output("jns %p",t); }

	void Assembler::CALL( S n, const NIns* t)       { emit_target32(n, X64_call, t); asm_output("call %p",t); }
	void Assembler::CALL64(S n)                     { emit_targetIpRel(n, X64_calli, 0); asm_output("call [ip]"); }
	void Assembler::RET()                           { emit(X64_ret);     asm_output("ret");         }

	void Assembler::MOVQMI(R r, I d, I32 imm)       { emitrm_imm32(X64_movqmi,r,d,imm); asm_output("movq %d(%s), %d",d,RQ(r),imm); }
	void Assembler::MOVLMI(R r, I d, I32 imm)       { emitrm_imm32(X64_movlmi,r,d,imm); asm_output("movl %d(%s), %d",d,RQ(r),imm); }
	void Assembler::MOVSMI(R r, I d, I32 imm)       { emitprm_imm16(X64_movsmi,r,d,imm); asm_output("movs %d(%s), %d",d,RQ(r),imm); }
	void Assembler::MOVBMI(R r, I d, I32 imm)       { emitrm_imm8(X64_movbmi,r,d,imm); asm_output("movb %d(%s), %d",d,RQ(r),imm); }

	void Assembler::MOVQSPR(I d, R r)               { emit(X64_movqspr | U64(d) << 56 | U64((REGNUM(r)&7)<<3) << 40 | U64((REGNUM(r)&8)>>1) << 24); asm_output("movq %d(rsp), %s", d, RQ(r)); }    // insert r into mod/rm and rex bytes
	void Assembler::MOVQSPX(I d, R r)               { emit(rexprb(X64_movqspx,RSP,r) | U64(d) << 56 | U64((REGNUM(r)&7)<<3) << 40); asm_output("movq %d(rsp), %s", d, RQ(r)); }

	void Assembler::XORPSA(R r, I32 i32)            { emitxm_abs(X64_xorpsa, r, i32); asm_output("xorps %s, (0x%x)",RQ(r), i32); }
	void Assembler::XORPSM(R r, NIns* a64)          { emitxm_rel(X64_xorpsm, r, a64); asm_output("xorps %s, (%p)", RQ(r), a64); }

	void Assembler::ANDPSA(R r, I32 i32)            { emitxm_abs(X64_andpsa, r, i32); asm_output("andps %s, (0x%x)",RQ(r), i32); }
	void Assembler::ANDPSM(R r, NIns* a64)          { emitxm_rel(X64_andpsm, r, a64); asm_output("andps %s, (%p)", RQ(r), a64); }

	void Assembler::X86_AND8R(R r)                  { emit(X86_and8r | U64(REGNUM(r)<<3|(REGNUM(r)|4))<<56); asm_output("andb %s, %s", RB(r), RBhi(r)); }
	void Assembler::X86_SETNP(R r)                  { emit(X86_setnp | U64(REGNUM(r)|4)<<56); asm_output("setnp %s", RBhi(r)); }
	void Assembler::X86_SETE(R r)                   { emit(X86_sete  | U64(REGNUM(r))<<56);   asm_output("sete %s", RB(r)); }

	void Assembler::XCHG(R r, R a)                  { emitrr(X64_xchgqrr, r, a); asm_output("xchg %s, %s", RQ(r), RQ(a)); }

#undef R
#undef I
#undef I32
#undef U64
#undef S

	void Assembler::MR(Register d, Register s) {
		NanoAssert(IsGpReg(d) && IsGpReg(s));
		MOVQR(d, s);
	}

	static bool isImm32(LIns *ins) {
		return ins->canOptimize() && (ins->isImmI() || (ins->isImmQ() && isS32(ins->immQ())));
	}
	static int32_t getImm32(LIns *ins) {
		return ins->isImmI() ? ins->immI() : int32_t(ins->immQ());
	}

	void Assembler::JMP(const NIns *target) {
		if (!target || isTargetWithinS32(target)) {
			JMP32(8, target);
		} else {
			// call underrunProtect before to prevent any page change during
			// when the ip relative jump is emitted.
			underrunProtect(16);

			_farJumpList->push_back(makePair(_nIns - 4, target));

			verbose_only(
				if (_logc->getLcBits() & LC_Native) {
					setOutputForEOL(" <= ip relative jump to %016x",
					target); } )
			JMP64(16);
		}
	}

	void Assembler::asm_shift(LIns *ins) {
		// Shift requires rcx for shift count.
		LIns *a = ins->oprnd1();
		LIns *b = ins->oprnd2();
		if (b->canOptimize() && isImm32(b)) {
			asm_shift_imm(ins);
			return;
		}

		Register rr, ra;
		if (a != b) {
			findSpecificRegFor(b, RCX);
			beginOp1Regs(ins, GpRegs & ~rmask(RCX), rr, ra);
		} else {
			// Nb: this is just like beginOp1Regs() except that it asserts
			// that ra is in GpRegs instead of rmask(RCX)) -- this is
			// necessary for the a==b case because 'a' might not be in RCX
			// (which is ok, the MR(rr, ra) below will move it into RCX).
			rr = prepareResultReg(ins, rmask(RCX));

			// If 'a' isn't in a register, it can be clobbered by 'ins'.
			ra = a->isInReg() ? a->getReg() : rr;
			NanoAssert(rmask(ra) & GpRegs);
		}

		switch (ins->opcode()) {
		default:
			TODO(asm_shift);
		case LIR_rshuq: SHRQ(rr);   break;
		case LIR_rshq:  SARQ(rr);   break;
		case LIR_lshq:  SHLQ(rr);   break;
		case LIR_rorq:  RORQ(rr);   break;
		case LIR_rcrq:  RCRQ(rr);   break;
		case LIR_rshui: SHRL(rr);   break;
		case LIR_rshi:  SARL(rr);   break;
		case LIR_lshi:  SHLL(rr);   break;
		case LIR_rori:  RORL(rr);   break;
		case LIR_rcri:  RCRL(rr);   break;
		}
		if (rr != ra)
			MR(rr, ra);

		endOpRegs(ins, rr, ra);
	}

	void Assembler::asm_shift_imm(LIns *ins) {
		Register rr, ra;
		beginOp1Regs(ins, GpRegs, rr, ra);

		LIns* opImm = ins->oprnd2();
		int shift = getImm32(opImm) & 63;

		switch (ins->opcode()) {
		default: TODO(shiftimm);
		case LIR_rshuq: SHRQI(rr, shift);   break;
		case LIR_rshq:  SARQI(rr, shift);   break;
		case LIR_lshq:  SHLQI(rr, shift);   break;
		case LIR_rorq:  RORQI(rr, shift);   break;
		case LIR_rcrq:  RCRQI(rr, shift);   break;
		case LIR_rshui: SHRLI( rr, shift);  break;
		case LIR_rshi:  SARLI( rr, shift);  break;
		case LIR_lshi:  SHLLI( rr, shift);  break;
		case LIR_rori:  RORLI( rr, shift);  break;
		case LIR_rcri:  RCRLI(rr, shift);   break;
		}
		if (rr != ra)
			MR(rr, ra);

		endOpRegs(ins, rr, ra);
	}

	// Binary op, integer regs, rhs is int32 constant.
	void Assembler::asm_arith_imm(LIns *ins) {
		LIns *b = ins->oprnd2();
		int32_t imm = getImm32(b);
		LOpcode op = ins->opcode();
		Register rr, ra;

		beginOp1Regs(ins, GpRegs, rr, ra);

		if (isS8(imm)) {
			switch (op) {
			default: NanoAssert(0);                 break;
			case LIR_addi:      ADDLR8(rr,  imm);   break;   // XXX: bug 547125: could use LEA for LIR_addi
			case LIR_addci:     ADCLR8(rr,  imm);   break;
			case LIR_andi:      ANDLR8(rr,  imm);   break;
			case LIR_ori:       ORLR8( rr,  imm);   break;
			case LIR_subi:      SUBLR8(rr,  imm);   break;
			case LIR_subbi:     SBBLR8(rr,  imm);   break;
			case LIR_xori:      XORLR8(rr,  imm);   break;
			case LIR_addq:      ADDQR8(rr,  imm);   break;
			case LIR_addcq:     ADCQR8(rr,  imm);   break;
			case LIR_subq:      SUBQR8(rr,  imm);   break;
			case LIR_subbq:     SBBQR8(rr,  imm);   break;
			case LIR_andq:      ANDQR8(rr,  imm);   break;
			case LIR_orq:       ORQR8( rr,  imm);   break;
			case LIR_xorq:      XORQR8(rr,  imm);   break;
			}
		} else {
			switch (op) {
			default: NanoAssert(0);                 break;
			case LIR_addi:      ADDLRI(rr,  imm);   break;   // XXX: bug 547125: could use LEA for LIR_addi
			case LIR_addci:     ADCLRI(rr,  imm);   break;
			case LIR_andi:      ANDLRI(rr,  imm);   break;
			case LIR_ori:       ORLRI( rr,  imm);   break;
			case LIR_subi:      SUBLRI(rr,  imm);   break;
			case LIR_subbi:     SBBLRI(rr,  imm);   break;
			case LIR_xori:      XORLRI(rr,  imm);   break;
			case LIR_addq:      ADDQRI(rr,  imm);   break;
			case LIR_addcq:     ADCQRI(rr,  imm);   break;
			case LIR_subq:      SUBQRI(rr,  imm);   break;
			case LIR_subbq:     SBBQRI(rr,  imm);   break;
			case LIR_andq:      ANDQRI(rr,  imm);   break;
			case LIR_orq:       ORQRI( rr,  imm);   break;
			case LIR_xorq:      XORQRI(rr,  imm);   break;
			}
		}
		if (rr != ra)
			MR(rr, ra);

		endOpRegs(ins, rr, ra);
	}

	// Generates code for a LIR_divi that doesn't have a subsequent LIR_modi.
	void Assembler::asm_div(LIns *div) {
		NanoAssert(div->isop(LIR_diviu) || div->isop(LIR_divis));
		LIns *a = div->oprnd1();
		LIns *b = div->oprnd2();

		evictIfActive(RDX);
		prepareResultReg(div, rmask(RAX));

		Register rb = findRegFor(b, GpRegs & ~(rmask(RAX)|rmask(RDX)));
		Register ra = a->isInReg() ? a->getReg() : RAX;

		IDIV(rb);
		SARLI(RDX, 31);
		MR(RDX, RAX);
		if (RAX != ra)
			MR(RAX, ra);

		freeResourcesOf(div);
		if (!a->isInReg()) {
			NanoAssert(ra == RAX);
			findSpecificRegForUnallocated(a, RAX);
		}
	}

	// Generates code for a LIR_modi(LIR_divi(divL, divR)) sequence.
	void Assembler::asm_div_mod(LIns *mod) {
		LIns *div = mod->oprnd1();

		NanoAssert(mod->isop(LIR_modi));
		NanoAssert(div->isop(LIR_diviu) || div->isop(LIR_divis));

		LIns *divL = div->oprnd1();
		LIns *divR = div->oprnd2();

		prepareResultReg(mod, rmask(RDX));
		prepareResultReg(div, rmask(RAX));

		Register rDivR = findRegFor(divR, GpRegs & ~(rmask(RAX)|rmask(RDX)));
		Register rDivL = divL->isInReg() ? divL->getReg() : RAX;

		IDIV(rDivR);
		SARLI(RDX, 31);
		MR(RDX, RAX);
		if (RAX != rDivL)
			MR(RAX, rDivL);

		freeResourcesOf(mod);
		freeResourcesOf(div);
		if (!divL->isInReg()) {
			NanoAssert(rDivL == RAX);
			findSpecificRegForUnallocated(divL, RAX);
		}
	}

	void Assembler::asm_mul(LIns* ins)
	{
		LOpcode     op      = ins->opcode();
		LIns*       oprnd2  = ins->oprnd2();

		Register rr, ra, rb;

		if (oprnd2->canOptimize() && oprnd2->isImmI())
		{
			int32_t imm = oprnd2->immI();
			// Special case: imul-by-imm has true 3-addr form.  So we don't
			// need the MR(rr, ra) after the IMULI.
			beginOp1Regs(ins, GpRegs, rr, ra);

			if (ins->getGenFlags() && (ins->flagRequested() != 0))
				asm_force_gen_flags(ins);

			switch (op)
			{
				case LIR_mullsi:
				case LIR_mullui:
				case LIR_muli:
					IMULI(rr, ra, imm);
					break;
				case LIR_mullsq:
				case LIR_mulluq:
				case LIR_mulq:
					IMULIQ(rr, ra, imm);
					break;

				default:                NanoAssert(0); // invalid opcode
			}

			endOpRegs(ins, rr, ra);

		}
		else
		{
			beginOp2Regs(ins, GpRegs, rr, ra, rb);

			if (ins->getGenFlags() && (ins->flagRequested() != 0))
				asm_force_gen_flags(ins);

			switch (op)
			{
				case LIR_mullsi:
				case LIR_mullui:
				case LIR_muli:
					IMUL(rr, rb);
					break;

				case LIR_mullsq:
				case LIR_mulluq:
				case LIR_mulq:
					IMULQ(rr, rb);
					break;

				default:
					NanoAssert(0); // invalid opcode
			}

			if (rr != ra)
				MR(rr, ra);

			endOpRegs(ins, rr, ra);

		}
	}

	// binary op with integer registers
	void Assembler::nArith(LIns *ins) {
		Register rr, ra, rb = UnspecifiedReg; // init to shut GCC up

		switch (ins->opcode()) {
		case LIR_lshi:  case LIR_lshq:
		case LIR_rshi:  case LIR_rshq:
		case LIR_rshui: case LIR_rshuq:
		case LIR_rori:  case LIR_rorq:
		case LIR_rcri: case LIR_rcrq:
			asm_shift(ins);
			return;
		case LIR_modi:
			asm_div_mod(ins);
			return;
		case LIR_divis:
		case LIR_diviu:
			// Nb: if the div feeds into a mod it will be handled by
			// asm_div_mod() rather than here.
			asm_div(ins);
			return; // !! return

		case LIR_muli:
		case LIR_mulq:
			asm_mul(ins);
			return; // !! return

		default:
			break;
		}

		LOpcode     op  = ins->opcode();
		LIns*       b   = ins->oprnd2();

		if (isImm32(b) && (op != LIR_andci) && (op != LIR_andcq) && b->canOptimize()) {
			asm_arith_imm(ins);
			return;
		}

		beginOp2Regs(ins, GpRegs, rr, ra, rb);
		switch (op) {
		default:            TODO(nArith);    break;
		case LIR_addi:      ADDRR(rr, rb);      break;  // XXX: bug 547125: could use LEA for LIR_addi
		case LIR_addci:     ADCRR(rr, rb);      break;
		case LIR_subi:      SUBRR(rr, rb);      break;
		case LIR_subbi:     SBBRR(rr, rb);      break;
		case LIR_andi:      ANDRR(rr, rb);      break;
		case LIR_andci:
		{
			Register regTemp = allocTempReg(GpRegs & ~(rmask(ra) | rmask(rb) | rmask(rr)));
			ANDRR(rr, regTemp);
			NOT(regTemp);
			MR(regTemp, rb);
			break;
		}
		case LIR_ori:       ORLRR(rr, rb);      break;
		case LIR_xori:      XORRR(rr, rb);      break;

		case LIR_addq:      ADDQRR(rr, rb);     break;
		case LIR_addcq:     ADCQRR(rr, rb);     break;
		case LIR_subq:      SUBQRR(rr, rb);     break;
		case LIR_subbq:     SBBQRR(rr, rb);     break;
		case LIR_andq:      ANDQRR(rr, rb);     break;
		case LIR_andcq:
		{
			Register regTemp = allocTempReg(GpRegs & ~(rmask(ra) | rmask(rb) | rmask(rr)));
			ANDQRR(rr, regTemp);
			NOTQ(regTemp);
			MR(regTemp, rb);
			break;
		}
		case LIR_orq:       ORQRR(rr, rb);      break;
		case LIR_xorq:      XORQRR(rr, rb);     break;
		}
		if (rr != ra)
			MR(rr, ra);

		endOpRegs(ins, rr, ra);
	}

	void Assembler::nMull(LIns* ins) {

		// nMullhi was not live
		// we don't need the high part
		// fallback on more flexible "imul r0, r1" instruction
		asm_mul(ins);
	}

	void Assembler::nMullhi(LIns* _ins) {

		LIns*       insLo = _ins->oprnd1();
		LIns*       insHi = _ins;
		LOpcode     op    = insLo->opcode();
		LIns*       a     = insLo->oprnd1();
		LIns*       b     = insLo->oprnd2();

		setResultLive(a);
		setResultLive(b);

		// result will be stored in EDX:EAX or RDX:RAX
		Register ra, rb;
		RegisterMask allow = GpRegs & ~(rmask(RAX) | rmask(RDX));
		
		ra = a->isInReg() ? findRegFor(a, allow) : RAX;
		if (a != b)
			rb = findRegFor(b, allow & ~rmask(ra));
		else
			rb = ra;

		prepareResultReg(insLo, rmask(RAX));
		prepareResultReg(insHi, rmask(RDX));

		NanoAssert(!(insHi->getGenFlags() && insLo->getGenFlags()));

		if (insLo->getGenFlags() && (insLo->flagRequested() != 0))
			asm_force_gen_flags(insLo);
		if (insHi->getGenFlags() && (insHi->flagRequested() != 0))
			asm_force_gen_flags(insHi);

		switch (op) {
			case LIR_mullui:			UMULL(rb); break;
			case LIR_mullsi:			SMULL(rb); break;
			case LIR_mulluq:			UMULLQ(rb); break;
			case LIR_mullsq:			SMULLQ(rb); break;
			default:					NanoAssert(0); break; // invalid opcode
		}

		if (RAX != ra)
			MR(RAX, ra);

		freeResourcesOf(insHi);
		freeResourcesOf(insLo);

		if (!a->isInReg())
			findSpecificRegForUnallocated(a, RAX);
	}

	// Binary op with fp registers.
	void Assembler::nFop(LIns *ins) {
		Register rr, ra, rb = UnspecifiedReg;   // init to shut GCC up
		beginOp2Regs(ins, FpRegs, rr, ra, rb);
		switch (ins->opcode()) {
		default:        TODO(nFop);
		case LIR_divd:  DIVSD(rr, rb); break;
		case LIR_muld:  MULSD(rr, rb); break;
		case LIR_addd:  ADDSD(rr, rb); break;
		case LIR_subd:  SUBSD(rr, rb); break;
		case LIR_divf:  DIVSS(rr, rb); break;
		case LIR_mulf:  MULSS(rr, rb); break;
		case LIR_addf:  ADDSS(rr, rb); break;
		case LIR_subf:  SUBSS(rr, rb); break;
		}
		if (rr != ra) {
			nNonGpCopy(rr, ra);
		}

		endOpRegs(ins, rr, ra);
	}

	void Assembler::nNegNot(LIns *ins) {
		NanoAssert(ins->isop(LIR_noti) || ins->isop(LIR_notq) || ins->isop(LIR_negi) || ins->isop(LIR_negq));

		Register rr, ra;
		beginOp1Regs(ins, GpRegs, rr, ra);

		LOpcode  op = ins->opcode();

		switch (op)
		{
			case LIR_noti:
			{
				if (ins->getGenFlags() && (ins->flagRequested() != 0))
					asm_force_gen_flags(ins);

				NOT(rr);
				break;
			}
			case LIR_notq:
			{
				if (ins->getGenFlags() && (ins->flagRequested() != 0))
					asm_force_gen_flags(ins);

				NOTQ(rr);
				break;
			}

			case LIR_negi:
				NEG(rr);
				break;
			case LIR_negq:
				NEGQ(rr);
				break;

			default:
				NanoAssert(0);
				break;
		}

		if (rr != ra)
			MR(rr, ra);

		endOpRegs(ins, rr, ra);
	}

	void Assembler::nCall(LIns* ins, const NIns *address) {

		ParameterValue*     callParams  = ins->callArgs();
		unsigned int        paramCount  = ins->callArgc();
		RegisterMask        evictMask   = ins->callVolatileMask();

		switch (ins->opcode())
		{
			case LIR_calli:
			CASE64(LIR_callq:)
			{
				Register r = prepareResultReg(ins, rmask(RAX));
				evictMask &= ~rmask(r);
				break;
			}

			case LIR_callf:
			case LIR_calld:
			{
				Register r = prepareResultReg(ins, rmask(XMM0));
				evictMask &= ~rmask(r);
				break;
			}

			case LIR_callv:
				break;

			default:
				NanoAssert(0);
		}

		evictSomeActiveRegs(evictMask);

		if ((address != 0) && isTargetWithinS32(address) && (_codeAlloc->absoluteOffset() == 0)) {
			CALL(8, address);
		} else {
			// call underrunProtect before to prevent any page change during
			// the emition of the ip relative jump
			underrunProtect(16);
			_farJumpList->push_back(makePair(_nIns - 4, address));

			verbose_only(
				RefBuf b;
				if (_logc->getLcBits() & LC_Native) {
					setOutputForEOL(" <= ip relative call to %016x (%s)",
						address,
						_printer->formatRef(&b, ins)); } )
			CALL64(16);
		}

		if (ins->opcode() != LIR_callv) {
			freeResourcesOf(ins);
		}

		reconcileParamRegs(callParams, paramCount);
	}

	void Assembler::nMovI(LIns *ins) {
		LOpcode op = ins->opcode();

		Register rr, ra;

		switch (op) {
			case LIR_movq2i:
				beginOp1Regs(ins, GpRegs, rr, ra);
				MOVSXDR(rr, ra);    // sign extend 32->64
				endOpRegs(ins, rr, ra);
				break;

			case LIR_movui2uq:
				beginOp1Regs(ins, GpRegs, rr, ra);
				if (ra != rr)
					MOVLR(rr, ra);
				endOpRegs(ins, rr, ra);
				break;

			case LIR_movf2i:
			case LIR_movd2i:
				rr = prepareResultReg(ins, GpRegs);
				ra = findRegFor(ins->oprnd1(), FpRegs);
				nNonGpCopy(rr, ra);
				freeResourcesOf(ins);
				break;

			default:
				NanoAssert(0); // unsupported opcode
				break;
		}


	}

	void Assembler::nMovQ(LIns *ins) {
		LOpcode op = ins->opcode();
		Register rr, ra;

		switch (op) {
			case LIR_movsi2sq:
				beginOp1Regs(ins, GpRegs, rr, ra);
				NanoAssert(IsGpReg(ra));
				MOVSXDR(rr, ra);    // sign extend 32->64
				endOpRegs(ins, rr, ra);
				break;

			case LIR_movui2uq:
				beginOp1Regs(ins, GpRegs, rr, ra);
				NanoAssert(IsGpReg(ra));
				if (ra != rr)
					MOVLR(rr, ra);
				endOpRegs(ins, rr, ra);
				break;

			case LIR_movd2q:
				rr = prepareResultReg(ins, GpRegs);
				ra = findRegFor(ins->oprnd1(), FpRegs);
				nNonGpCopy(rr, ra);
				freeResourcesOf(ins);
				break;

			default:
				NanoAssert(0); // invalid opcode
				break;
		}
	}

	void Assembler::nMovF(LIns* ins) {
		LOpcode op = ins->opcode();
		Register rr, ra;
		
		switch (op) {

			case LIR_movi2f:
			case LIR_movd2f:
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(ins->oprnd1(), GpRegs);
				nNonGpCopy(rr, ra);
				freeResourcesOf(ins);
				break;

			default:
				NanoAssert(0); // invalid opcode
				break;
		}

	}

	void Assembler::nMovD(LIns* ins) {
		LOpcode op = ins->opcode();
		Register rr, ra;

		switch (op) {
			case LIR_movf2d:
			case LIR_movi2d:
			case LIR_movq2d:
				beginOp1Regs(ins, FpRegs, rr, ra);
				if (ra != rr)
					nNonGpCopy(rr, ra);
				endOpRegs(ins, rr, ra);
				break;

			default:
				NanoAssert(0); // invalid opcode
				break;
		}
	}

	void Assembler::nFconv(LIns* ins) {
		LOpcode op = ins->opcode();
		LIns *a = ins->oprnd1();
		Register rr, ra, rtmp;

		// The CVTSI2SD instruction only writes to the low 64bits of the target
		// XMM register, which hinders register renaming and makes dependence
		// chains longer.  So we precede with XORPS to clear the target register.


		switch (op) {

			case LIR_cvtif:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTPS2DQ(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtuif:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				MOVQXR(rr, rtmp);
				CVTSS2SQ(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtitf:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTTPS2DQ(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtuitf:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				MOVQXR(rr, rtmp);
				CVTTSS2SQ(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvfif:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTDQ2PS(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvfuif:
				NanoAssert(ins->isF() && a->isF());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				CVTSQ2SS(rr, rtmp);
				MOVQRX(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtid:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTPD2DQ(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtuid:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				MOVQXR(rr, rtmp);
				CVTSD2SQ(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtitd:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTTPD2DQ(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvtuitd:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				MOVQXR(rr, rtmp);
				CVTTSD2SQ(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvfid:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				CVTSI2SD(rr, rtmp);
				MOVQRX(rtmp, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvfuid:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				rtmp = allocTempReg(GpRegs);
				ra = findRegFor(a, FpRegs);
				CVTSQ2SD(rr, rtmp);
				MOVQRX(rtmp, ra);
				freeResourcesOf(ins);
				break;

			//case LIR_cvtqd:
			//	break;
			//
			//case LIR_cvtuqd:
			//	break;
			//
			//case LIR_cvtqtd:
			//	break;
			//
			//case LIR_cvtuqtd:
			//	break;

			case LIR_cvfqd:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, GpRegs);
				CVTSQ2SD(rr, ra);   // cvtsq2sd xmmr, b  only writes xmm:0:64
				XORPS(rr, rr);          // xorps xmmr,xmmr to break dependency chains
				freeResourcesOf(ins);
				break;

			//case LIR_cvfuqd:
			//	break;

			case LIR_cvtfd:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTSD2SS(rr, ra);
				freeResourcesOf(ins);
				break;

			case LIR_cvffd:
				NanoAssert(ins->isD() && a->isD());
				rr = prepareResultReg(ins, FpRegs);
				ra = findRegFor(a, FpRegs);
				CVTSS2SD(rr, ra);
				freeResourcesOf(ins);
				break;

			default:
				NanoAssert(0); // invalid opcode
				break;
		}
	}


	void Assembler::nCmov(LIns *ins) {
		LIns* cond    = ins->oprnd1();
		LIns* iftrue  = ins->oprnd2();
		LIns* iffalse = ins->oprnd3();
		NanoAssert(cond->isCmp());
		NanoAssert(iftrue->retType() == ins->retType());
		NanoAssert(iftrue->retType() == iffalse->retType());

#if defined(NANOJIT_ENABLE_ASSERT)
		checkCondition(cond);
#endif

		bool isFloatOp		= ins->isForD();

		if ( isFloatOp ) {
			LTy condType = cond->oprnd1()->retType();
			if ((condType == LTy_F) || (condType == LTy_D)) {
				emit_cmovf(ins);
			} else {
				RegisterMask allow = isFloatOp ? FpRegs : GpRegs;
				Register rr = prepareResultReg(ins, allow);
				Register rf = findRegFor(iffalse, allow & ~rmask(rr));

				NIns* target = _nIns;
				nNonGpCopy(rr, rf);
				Branches branches;
				nRelativeBranchCond(false, cond, target, false, &branches);

				// If 'iftrue' isn't in a register, it can be clobbered by 'ins'.
				Register rt = iftrue->isInReg() ? iftrue->getReg() : rr;

				if (rr != rt)
					nNonGpCopy(rr, rt);

				freeResourcesOf(ins);
				if (!iftrue->isInReg()) {
					NanoAssert(rt == rr);
					findSpecificRegForUnallocated(iftrue, rr);
				}
				asm_cmp(cond);
			}
			return;
		}

		RegisterMask allow = GpRegs;
		Register rr = prepareResultReg(ins, allow);
		Register rf = findRegFor(iffalse, allow & ~rmask(rr));

		// If 'iftrue' isn't in a register, it can be clobbered by 'ins'.
		Register rt = iftrue->isInReg() ? iftrue->getReg() : rr;

		// WARNING: We cannot generate any code that affects the condition
		// codes between the MRcc generation here and the asm_cmpi() call
		// below.  See asm_cmpi() for more details.
		LOpcode condop = cond->opcode();
		if (ins->isop(LIR_cmovi)) {
			switch (condop) {
			case LIR_eqi:  case LIR_eqq:  case LIR_eqf: case LIR_eqd:	CMOVNE (rr, rf);    break;
			case LIR_nei:  case LIR_neq:  case LIR_nef: case LIR_ned:	CMOVE  (rr, rf);    break;
			case LIR_lti:  case LIR_ltq:								CMOVNL (rr, rf);    break;
			case LIR_gti:  case LIR_gtq:								CMOVNG (rr, rf);    break;
			case LIR_lei:  case LIR_leq:								CMOVNLE(rr, rf);    break;
			case LIR_gei:  case LIR_geq:								CMOVNGE(rr, rf);    break;
			case LIR_ltui: case LIR_ltuq: case LIR_ltf:	case LIR_ltd:	CMOVNB (rr, rf);    break;
			case LIR_gtui: case LIR_gtuq: case LIR_gtf: case LIR_gtd:	CMOVNA (rr, rf);    break;
			case LIR_leui: case LIR_leuq: case LIR_lef:	case LIR_led:	CMOVNBE(rr, rf);    break;
			case LIR_geui: case LIR_geuq: case LIR_gef:	case LIR_ged:	CMOVNAE(rr, rf);    break;
			case LIR_unf:  case LIR_und:								CMOVNP (rr, rf);    break;
			default:                        NanoAssert(0);  break;
			}
		} else {
			NanoAssert(ins->isop(LIR_cmovq));
			switch (condop) {
			case LIR_eqi:  case LIR_eqq:    CMOVQNE( rr, rf); break;
			case LIR_nei:  case LIR_neq:    CMOVQE( rr, rf);  break;
			case LIR_lti:  case LIR_ltq:    CMOVQNL( rr, rf); break;
			case LIR_gti:  case LIR_gtq:    CMOVQNG( rr, rf); break;
			case LIR_lei:  case LIR_leq:    CMOVQNLE(rr, rf); break;
			case LIR_gei:  case LIR_geq:    CMOVQNGE(rr, rf); break;
			case LIR_ltui: case LIR_ltuq:   CMOVQNB( rr, rf); break;
			case LIR_gtui: case LIR_gtuq:   CMOVQNA( rr, rf); break;
			case LIR_leui: case LIR_leuq:   CMOVQNBE(rr, rf); break;
			case LIR_geui: case LIR_geuq:   CMOVQNAE(rr, rf); break;
			default:                        NanoAssert(0);    break;
			}
		}
		if (rr != rt)
			MR(rr, rt);

		freeResourcesOf(ins);
		if (!iftrue->isInReg()) {
			NanoAssert(rt == rr);
			findSpecificRegForUnallocated(iftrue, rr);
		}
		asm_cmp(cond);
	}

	NIns* Assembler::nRelativeBranch(const NIns* target) {
		JMP(target);
		return _nIns;
	}

	void Assembler::nRelativeBranchCond(bool onFalse, LIns *cond, const NIns *target, bool patchable, Branches* branches) {
		if (target && !isTargetWithinS32(target)) {
			// A conditional jump beyond 32-bit range, so invert the
			// branch/compare and emit an unconditional jump to the target:
			//         j(inverted) B1
			//         jmp target
			//     B1:
			NIns* shortTarget = _nIns;
			JMP(target);
			target = shortTarget;
			onFalse = !onFalse;

			branches->addBranch(_nIns, shortTarget);
		}

		// float4 is handled by the "branchi" path
		if (isCmpDOpcode(cond->opcode()) || isCmpFOpcode(cond->opcode()))
			 asm_branchd_helper(onFalse, cond, target, patchable, branches);
		else
			 asm_branchi_helper(onFalse, cond, target, patchable, branches);

		asm_cmp(cond);
	}

	void Assembler::asm_branchi_helper(bool onFalse, LIns *cond, const NIns *target, bool patchable, Branches* branches) {
		// We must ensure there's room for the instruction before calculating
		// the offset. And the offset determines the opcode (8bit or 32bit).
		LOpcode condop = cond->opcode();

		if (target && (!patchable) && isTargetWithinS8(target)) {
			if (onFalse) {
				switch (condop) {
				case LIR_eqi: case LIR_eqq:         JNE8( 8, target);   break;
				case LIR_nei: case LIR_neq:         JE8(  8, target);   break;
				case LIR_lti: case LIR_ltq:         JNL8( 8, target);   break;
				case LIR_gti: case LIR_gtq:         JNG8( 8, target);   break;
				case LIR_lei: case LIR_leq:         JNLE8(8, target);   break;
				case LIR_gei: case LIR_geq:         JNGE8(8, target);   break;
				case LIR_ltui: case LIR_ltuq:       JNB8( 8, target);   break;
				case LIR_gtui: case LIR_gtuq:       JNA8( 8, target);   break;
				case LIR_leui: case LIR_leuq:       JNBE8(8, target);   break;
				case LIR_geui: case LIR_geuq:       JNAE8(8, target);   break;
				default:                            NanoAssert(0);      break;
				}
			} else {
				switch (condop) {
				case LIR_eqi:  case LIR_eqq:        JE8( 8, target);    break;
				case LIR_nei:  case LIR_neq:        JNE8(8, target);    break;
				case LIR_lti:  case LIR_ltq:        JL8( 8, target);    break;
				case LIR_gti:  case LIR_gtq:        JG8( 8, target);    break;
				case LIR_lei:  case LIR_leq:        JLE8(8, target);    break;
				case LIR_gei:  case LIR_geq:        JGE8(8, target);    break;
				case LIR_ltui: case LIR_ltuq:       JB8( 8, target);    break;
				case LIR_gtui: case LIR_gtuq:       JA8( 8, target);    break;
				case LIR_leui: case LIR_leuq:       JBE8(8, target);    break;
				case LIR_geui: case LIR_geuq:       JAE8(8, target);    break;
				default:                            NanoAssert(0);      break;
				}
			}
		} else {
			if (onFalse) {
				switch (condop) {
				case LIR_eqi:  case LIR_eqq:        JNE( 8, target);    break;
				case LIR_nei:  case LIR_neq:        JE(  8, target);    break;
				case LIR_lti:  case LIR_ltq:        JNL( 8, target);    break;
				case LIR_gti:  case LIR_gtq:        JNG( 8, target);    break;
				case LIR_lei:  case LIR_leq:        JNLE(8, target);    break;
				case LIR_gei:  case LIR_geq:        JNGE(8, target);    break;
				case LIR_ltui: case LIR_ltuq:       JNB( 8, target);    break;
				case LIR_gtui: case LIR_gtuq:       JNA( 8, target);    break;
				case LIR_leui: case LIR_leuq:       JNBE(8, target);    break;
				case LIR_geui: case LIR_geuq:       JNAE(8, target);    break;
				default:                            NanoAssert(0);      break;
				}
			} else {
				switch (condop) {
				case LIR_eqi:  case LIR_eqq:        JE( 8, target);     break;
				case LIR_nei:  case LIR_neq:        JNE(8,  target);    break;
				case LIR_lti:  case LIR_ltq:        JL(8, target);      break;
				case LIR_gti:  case LIR_gtq:        JG( 8, target);     break;
				case LIR_lei:  case LIR_leq:        JLE(8, target);     break;
				case LIR_gei:  case LIR_geq:        JGE(8, target);     break;
				case LIR_ltui: case LIR_ltuq:       JB( 8, target);     break;
				case LIR_gtui: case LIR_gtuq:       JA( 8, target);     break;
				case LIR_leui: case LIR_leuq:       JBE(8, target);     break;
				case LIR_geui: case LIR_geuq:       JAE(8, target);     break;
				default:                            NanoAssert(0);      break;
				}
			}
		}

		branches->addBranch(_nIns, target);
	}

	void Assembler::asm_cmp(LIns *cond) {
	  if (isCmpFOpcode(cond->opcode()) || isCmpDOpcode(cond->opcode()))
		  asm_cmpd(cond);
	  else
		  asm_cmpi(cond);
	}

	// WARNING: this function cannot generate code that will affect the
	// condition codes prior to the generation of the test/cmp. See
	// Nativei386.cpp:asm_cmpi() for details.
	void Assembler::asm_cmpi(LIns *cond) {

		LIns *b = cond->oprnd2();
		if (b->canOptimize() && isImm32(b)) {
			asm_cmpi_imm(cond);
			return;
		}
		LIns *a = cond->oprnd1();

		setResultLive(a);
		setResultLive(b);

		Register ra, rb;
		if (a != b) {
			findRegFor2(GpRegs, a, ra, GpRegs, b, rb);
		} else {
			// optimize-me: this will produce a const result!
			ra = rb = findRegFor(a, GpRegs);
		}

		LOpcode condop = cond->opcode();
		if (isCmpQOpcode(condop)) {
			CMPQR(ra, rb);
		} else {
			NanoAssert(isCmpIOpcode(condop));
			CMPLR(ra, rb);
		}
	}

	void Assembler::asm_cmpi_imm(LIns *cond) {
		LOpcode condop = cond->opcode();
		LIns *a = cond->oprnd1();
		LIns *b = cond->oprnd2();

		setResultLive(a);
		Register    ra      = findRegFor(a, GpRegs);
		int32_t     imm     = getImm32(b);
		if (isCmpQOpcode(condop)) {
			if (isS8(imm))
				CMPQR8(ra, imm);
			else
				CMPQRI(ra, imm);
		} else {
			NanoAssert(isCmpIOpcode(condop));
			if (isS8(imm))
				CMPLR8(ra, imm);
			else
				CMPLRI(ra, imm);
		}
	}

	// Compiling floating point branches.
	// Discussion in https://bugzilla.mozilla.org/show_bug.cgi?id=443886.
	//
	//  fucom/p/pp: c3 c2 c0   jae ja    jbe jb je jne
	//  ucomisd:     Z  P  C   !C  !C&!Z C|Z C  Z  !Z
	//              -- -- --   --  ----- --- -- -- --
	//  unordered    1  1  1             T   T  T
	//  greater >    0  0  0   T   T               T
	//  less    <    0  0  1             T   T     T
	//  equal   =    1  0  0   T         T      T
	//
	//  Here are the cases, using conditionals:
	//
	//  branch  >=  >   <=       <        =
	//  ------  --- --- ---      ---      ---
	//  LIR_jt  jae ja  swap+jae swap+ja  jp over je
	//  LIR_jf  jb  jbe swap+jb  swap+jbe jne+jp

	void Assembler::asm_branchd_helper(bool onFalse, LIns *cond, const NIns *target, bool patchable, Branches* branches) {
		NanoUnusedVariable(patchable);

		LOpcode condop = cond->opcode();
		if (isCmpFOpcode(condop))
			condop = getCmpDOpcode(condop);
		if (condop == LIR_eqd) {
			if (onFalse) {
				// branch if unordered or !=
				JP(16, target);     // underrun of 12 needed, round up for overhang --> 16
				branches->addBranch(_nIns, target);

				JNE(0, target);     // no underrun needed, previous was enough
				branches->addBranch(_nIns, target);

			} else {
				// jp skip (2byte)
				// jeq target
				// skip: ...
				underrunProtect(16); // underrun of 7 needed but we write 2 instr --> 16
				NIns *skip = _nIns;
				JE(0, target);      // no underrun needed, previous was enough
				branches->addBranch(_nIns, target);

				JP8(0, skip);       // ditto
				branches->addBranch(_nIns, skip);
			}
		}
		else {
			switch (condop) {
			case LIR_ltd: if (onFalse) JAE(8, target); else JB(8, target);  break;
			case LIR_gtd: if (onFalse) JBE(8, target); else JA(8, target);  break;
			case LIR_led: if (onFalse) JA(8, target);  else JBE(8, target); break;
			case LIR_ged: if (onFalse) JB(8, target);  else JAE(8, target); break;
			default:      NanoAssert(0);                                    break;
			}

			branches->addBranch(_nIns, target);
		}
	}

	// WARNING: This function cannot generate any code that will affect the
	// condition codes prior to the generation of the ucomisd/ucomiss.
	// See asm_cmpi() for more details.
	void Assembler::asm_cmpd(LIns *cond) {
		LOpcode opcode = cond->opcode();

		bool singlePrecision = isCmpFOpcode(opcode);

		LIns* a = cond->oprnd1();
		LIns* b = cond->oprnd2();

		setResultLive(a);
		setResultLive(b);

		Register ra, rb;
		findRegFor2(FpRegs, a, ra, FpRegs, b, rb);
		if (singlePrecision)
			UCOMISS(ra, rb);
		else
			UCOMISD(ra, rb);
	}

	// WARNING: the code generated by this function must not affect the
	// condition codes.  See asm_cmpi() for details.
	void Assembler::nRestore(LIns *ins, Register r) {

		// do not rematerialize immediate that cannot be optimized away
		if (ins->canOptimize())
		{
			if (ins->isImmI()) {
				nImmI(r, ins->immI(), /*canClobberCCs*/false);
				return;
			}
			else if (ins->isImmQ()) {
				nImmQ(r, ins->immQ(), /*canClobberCCs*/false);
				return;
			}
			else if (ins->isImmD()) {
				nImmD(r, ins->immDasQ(), /*canClobberCCs*/false);
				return;
			}
			else if (ins->isImmF()) {
				nImmF(r, ins->immFasI(), /*canClobberCCs*/false);
				return;
			}
		}

		NanoAssert(!RegAlloc::nCanRemat(ins));

		int d = findMemFor(ins);
		LTy retType = ins->retType();

		switch (retType)
		{
			case LTy_I:
				NanoAssert(ins->isI());
				switch (ins->getArSize()) {
					case 4:		MOVLRM(r, d, FP); break;
					case 2:		MOVZX16M(r, d, FP); break;
					case 1:		MOVZX8M(r, d, FP); break;
					default:	NanoAssert(0); break;
				}
				break;

			case LTy_Q:
				NanoAssert(IsGpReg(r));
				MOVQRM(r, d, FP);
				break;

			case LTy_F:
				NanoAssert(IsFpReg(r));
				MOVSSRM(r, d, FP);
				break;

			case LTy_D:
				NanoAssert(IsFpReg(r));
				MOVSDRM(r, d, FP);
				break;

			case LTy_V:
				NanoAssert(0);
				break;
		}
	}

	void Assembler::nGpCopy(Register d, Register s) {
		MR(d, s);
	}

	void Assembler::nNonGpCopy(Register d, Register s) {
		if (!IsFpReg(d) && IsFpReg(s)) {
			// gpr <- xmm: use movq r/m64, xmm (66 REX.W 0F 7E /r)
			MOVQRX(d, s);
		} else if (IsFpReg(d) && IsFpReg(s)) {
			// xmm <- xmm: use movaps. movsd r,r causes partial register stall
			MOVAPSR(d, s);
		} else {
			NanoAssert(IsFpReg(d) && !IsFpReg(s));
			// xmm <- gpr: use movq xmm, r/m64 (66 REX.W 0F 6E /r)
			MOVQXR(d, s);
		}
	}

	// Register setup for load ops.  Pairs with endLoadRegs().
	void Assembler::beginLoadRegs(LIns *ins, RegisterMask allow, Register &rr, int32_t &dr, Register &rb) {
		dr = ins->disp();
		LIns *base = ins->oprnd1();
		rb = getBaseReg(base, dr, BaseRegs);
		rr = prepareResultReg(ins, allow & ~rmask(rb));
	}

	// Register clean-up for load ops.  Pairs with beginLoadRegs().
	void Assembler::endLoadRegs(LIns* ins) {
		freeResourcesOf(ins);
	}

	void Assembler::nLoad64(LIns *ins) {
		Register rr, rb;
		int32_t dr;
		switch (ins->opcode()) {
			case LIR_ldq:
				beginLoadRegs(ins, GpRegs, rr, dr, rb);
				NanoAssert(IsGpReg(rr));
				MOVQRM(rr, dr, rb);     // general 64bit load, 32bit const displacement
				break;
			case LIR_ldd:
				beginLoadRegs(ins, FpRegs, rr, dr, rb);
				NanoAssert(IsFpReg(rr));
				MOVSDRM(rr, dr, rb);    // load 64bits into XMM
				break;
			case LIR_ldf:
				beginLoadRegs(ins, FpRegs, rr, dr, rb);
				NanoAssert(IsFpReg(rr));
				MOVSSRM(rr, dr, rb);
				break;
			case LIR_ldf2d:
				beginLoadRegs(ins, FpRegs, rr, dr, rb);
				NanoAssert(IsFpReg(rr));
				CVTSS2SD(rr, rr);
				MOVSSRM(rr, dr, rb);
				break;
			default:
				NanoAssertMsg(0, "nLoad64 should never receive this LIR opcode");
				break;
		}
		endLoadRegs(ins);
	}


	void Assembler::nLoad32(LIns *ins) {
		NanoAssert(ins->isI());
		Register r, b;
		int32_t d;
		beginLoadRegs(ins, GpRegs, r, d, b);
		LOpcode op = ins->opcode();
		switch (op) {
			case LIR_lduc2ui:
				MOVZX8M( r, d, b);
				break;
			case LIR_ldus2ui:
				MOVZX16M(r, d, b);
				break;
			case LIR_ldi:
				MOVLRM(  r, d, b);
				break;
			case LIR_ldc2i:
				MOVSX8M( r, d, b);
				break;
			case LIR_lds2i:
				MOVSX16M( r, d, b);
				break;
			default:
				NanoAssertMsg(0, "nLoad32 should never receive this LIR opcode");
				break;
		}
		endLoadRegs(ins);
	}

	void Assembler::nImmF(Register r, uint32_t v, bool canClobberCCs) {
		NanoAssert(IsFpReg(r));
		if (v == 0 && canClobberCCs) {
			XORPS(r, r);
		} else {
			// There's no general way to load an immediate into an XMM reg.
			// For non-zero floats the best thing is to put the equivalent
			// 64-bit integer into a scratch GpReg and then move it into the
			// appropriate FpReg.
			Register rt = allocTempReg(GpRegs);
			MOVDXR(r, rt);
			nImmI(rt, v, canClobberCCs);
		}
	}

	void Assembler::nStoreToStack(LIns* ins, int d, RegisterMask allow) {
		Register r = findRegFor(ins, GpRegs & allow);

		switch (ins->retType())
		{
			case LTy_I:
			case LTy_Q:
				NanoAssert((d % 8) == 0);
				MOVQSPR(d, r);
				break;

			default:
				NanoAssert(0); // invalid data type
		}

	}

	void Assembler::nStore64(LOpcode op, LIns *value, int d, LIns *base) {
		// This function also handles stf (store-float-32) because its more
		// convenient to do it here than nStore32, which only handles GP registers.
		NanoAssert(op == LIR_stf ? value->isF() : value->isQorD());

		switch (op) {
			case LIR_stq: {
				uint64_t c;
				if (value->canOptimize() && value->isImmQ() && (c = value->immQ(), isS32(c))) {
					uint64_t c = value->immQ();
					Register rb = getBaseReg(base, d, BaseRegs);
					// MOVQMI takes a 32-bit integer that gets signed extended to a 64-bit value.
					MOVQMI(rb, d, int32_t(c));
				} else {
					Register rr, rb;
					getBaseReg2(GpRegs, value, rr, BaseRegs, base, rb, d);
					MOVQMR(rr, d, rb);    // gpr store
				}
				break;
			}
			case LIR_std: {
				Register b = getBaseReg(base, d, BaseRegs);
				Register r = findRegFor(value, FpRegs);
				MOVSDMR(r, d, b);   // xmm store
				break;
			}
			case LIR_stf:{
				if (value->canOptimize() && value->isImmF()) {
					// use the integer pipeline to store the value
					Register rb = getBaseReg(base, d, BaseRegs);
					int c = value->immFasI();
					MOVLMI(rb, d, c);
				} else {
					Register b = getBaseReg(base, d, BaseRegs);
					Register r = findRegFor(value, FpRegs);
					MOVSSMR(r, d, b);   // store
				}
				break;
			}
			case LIR_std2f: {
				Register b = getBaseReg(base, d, BaseRegs);
				Register r = findRegFor(value, FpRegs);
				Register t = allocTempReg(FpRegs    & ~rmask(r));

				MOVSSMR(t, d, b);   // store
				CVTSD2SS(t, r);     // cvt to single-precision
				XORPS(t, t);           // break dependency chains
				break;
			}
			default:
				NanoAssertMsg(0, "nStore64 should never receive this LIR opcode");
				break;
		}
	}

	void Assembler::nStore32(LOpcode op, LIns *value, int d, LIns *base) {
		if (value->canOptimize() && value->isImmI()) {
			Register rb = getBaseReg(base, d, BaseRegs);
			int c = value->immI();
			switch (op) {
				case LIR_sti2c: MOVBMI(rb, d, c);   break;
				case LIR_sti2s: MOVSMI(rb, d, c);   break;
				case LIR_sti:   MOVLMI(rb, d, c);   break;
				default:        NanoAssert(0);      break;
			}

		} else {
			// Quirk of x86-64: reg cannot appear to be ah/bh/ch/dh for
			// single-byte stores with REX prefix.
			const RegisterMask SrcRegs = (op == LIR_sti2c) ? SingleByteStoreRegs : GpRegs;

			NanoAssert(value->isI());
			Register b = getBaseReg(base, d, BaseRegs);
			Register r = findRegFor(value, SrcRegs & ~rmask(b));

			switch (op) {
				case LIR_sti2c:     MOVBMR(r, d, b);    break;
				case LIR_sti2s:     MOVSMR(r, d, b);    break;
				case LIR_sti:       MOVLMR(r, d, b);    break;
				default:            NanoAssert(0);      break;
			}
		}
	}

	void Assembler::nImmI(LIns *ins) {
		Register rr = prepareResultReg(ins, GpRegs);

		if (ins->getGenFlags() && (ins->flagRequested() != 0)) {
			asm_force_gen_flags(ins);
		}

		nImmI(rr, ins->immI(), /*canClobberCCs*/true);
		freeResourcesOf(ins);
	}

	void Assembler::nImmQ(LIns *ins) {
		Register rr = prepareResultReg(ins, GpRegs);

		if (ins->getGenFlags() && (ins->flagRequested() != 0)) {
			asm_force_gen_flags(ins);
		}

		nImmQ(rr, ins->immQ(), /*canClobberCCs*/true);
		freeResourcesOf(ins);
	}

	void Assembler::nImmD(LIns *ins) {
		Register r = prepareResultReg(ins, FpRegs);
		nImmD(r, ins->immDasQ(), /*canClobberCCs*/true);
		freeResourcesOf(ins);
	}

	void Assembler::nImmF(LIns *ins) {
		Register r = prepareResultReg(ins, FpRegs);
		nImmF(r, ins->immFasI(), /*canClobberCCs*/true);
		freeResourcesOf(ins);
	}

	void Assembler::nImmI(Register r, int32_t v, bool canClobberCCs) {
		NanoAssert(IsGpReg(r));
		if (v == 0 && canClobberCCs) {
			XORRR(r, r);
		} else {
			MOVI(r, v);
		}
	}

	void Assembler::nImmQ(Register r, uint64_t v, bool canClobberCCs) {
		NanoAssert(IsGpReg(r));
		if (isU32(v)) {
			nImmI(r, int32_t(v), canClobberCCs);
		} else if (isS32(v)) {
			// safe for sign-extension 32->64
			MOVQI32(r, int32_t(v));
		// do not use LEARIP(r, d). Because of cache restores, the code must be position
		// independant, or must make sure it can be always patched.
		} else {
			MOVQI(r, v);
		}
	}

	void Assembler::nImmD(Register r, uint64_t v, bool canClobberCCs) {
		NanoAssert(IsFpReg(r));
		if (v == 0 && canClobberCCs) {
			XORPS(r, r);
		} else {
			// There's no general way to load an immediate into an XMM reg.
			// For non-zero floats the best thing is to put the equivalent
			// 64-bit integer into a scratch GpReg and then move it into the
			// appropriate FpReg.
			Register rt = allocTempReg(GpRegs);
			MOVQXR(r, rt);
			nImmQ(rt, v, canClobberCCs);
		}
	}

	// Register setup for 2-address style unary ops of the form R = (op) R.
	// Pairs with endOpRegs().
	void Assembler::beginOp1Regs(LIns* ins, RegisterMask allow, Register &rr, Register &ra) {
		LIns* a = ins->oprnd1();

		rr = prepareResultReg(ins, allow);

		// If 'a' isn't in a register, it can be clobbered by 'ins'.
		ra = a->isInReg() ? a->getReg() : rr;

		NanoAssert(rmask(ra) & allow);
	}

	// Register setup for 2-address style binary ops of the form R = R (op) B.
	// Pairs with endOpRegs().
	void Assembler::beginOp2Regs(LIns *ins, RegisterMask allow, Register &rr, Register &ra,
								 Register &rb) {
		LIns *a = ins->oprnd1();
		LIns *b = ins->oprnd2();
		if (a != b) {
			rb = findRegFor(b, allow);
			allow &= ~rmask(rb);
		}
		rr = prepareResultReg(ins, allow);

		// If 'a' isn't in a register, it can be clobbered by 'ins'.
		ra = a->isInReg() ? a->getReg() : rr;

		NanoAssert(rmask(ra) & allow);

		if (a == b) {
			rb = ra;
		}
	}

	// Register clean-up for 2-address style unary ops of the form R = (op) R.
	// Pairs with beginOp1Regs() and beginOp2Regs().
	void Assembler::endOpRegs(LIns* ins, Register rr, Register ra) {
		(void) rr; // quell warnings when NanoAssert is compiled out.

		LIns* a = ins->oprnd1();

		// We're finished with 'ins'.
		NanoAssert(ins->getReg() == rr);
		freeResourcesOf(ins);

		// If 'a' isn't in a register yet, that means it's clobbered by 'ins'.
		if (!a->isInReg()) {
			NanoAssert(ra == rr);
			findSpecificRegForUnallocated(a, ra);
		}
	}

	static const NANOJIT_ALIGN16(int64_t) negateMaskD[]  = { static_cast<int64_t>(0x8000000000000000), 0 };
	static const NANOJIT_ALIGN16(int32_t) negateMaskF[]  = { static_cast<int32_t>(0x80000000), 0, 0, 0 };
	static const NANOJIT_ALIGN16(int32_t) negateMaskF4[] = { static_cast<int32_t>(0x80000000), static_cast<int32_t>(0x80000000), static_cast<int32_t>(0x80000000), static_cast<int32_t>(0x80000000) };
	static const NANOJIT_ALIGN16(int32_t) absMaskF[]  = { static_cast<int32_t>(0x7FFFFFFF), 0, 0, 0 };
	static const NANOJIT_ALIGN16(int64_t) absMaskD[]  = { static_cast<int64_t>(0x7FFFFFFFFFFFFFFF), 0 };

	void Assembler::nNegAbs(LIns *ins) {
		Register rr, ra;
		NanoAssert(ins->isop(LIR_negf) || ins->isop(LIR_negd) || ins->isop(LIR_absf) || ins->isop(LIR_absd));
		beginOp1Regs(ins, FpRegs, rr, ra);
		uintptr_t mask;

		switch (ins->opcode()) {
		default: NanoAssert(!"bad opcode for nNegAbs"); mask = 0; break;
		case LIR_negf:    mask = (uintptr_t) negateMaskF;  break;
		case LIR_negd:    mask = (uintptr_t) negateMaskD;  break;
		case LIR_absf:    mask = (uintptr_t) absMaskF;     break;
		case LIR_absd:    mask = (uintptr_t) absMaskD;     break;
		}

		if (isS32(mask)) {
			// builtin code is in bottom or top 2GB addr space, use absolute addressing
			if (ins->isop(LIR_absf) || ins->isop(LIR_absd)) {
				ANDPSA(rr, (int32_t)mask);
			} else {
				XORPSA(rr, (int32_t)mask);
			}

		} else if (isTargetWithinS32((NIns*)mask)) {
			// jit code is within +/-2GB of builtin code, use rip-relative
			if (ins->isop(LIR_absf) || ins->isop(LIR_absd)) {
				ANDPSM(rr, (NIns*)mask);
			} else {
				XORPSM(rr, (NIns*)mask);
			}
		} else {
			// This is just hideous - can't use RIP-relative load, can't use
			// absolute-address load, and cant move imm64 const to XMM.
			// Solution: move negateMaskD into a temp GP register, then copy to
			// a temp XMM register.
			// Nb: we don't want any F64 values to end up in a GpReg, nor any
			// I64 values to end up in an FpReg.
			//
			//   # 'gt' and 'ga' are temporary GpRegs.
			//   # ins->oprnd1() is in 'rr' (FpRegs)
			//   mov   gt, 0x8000000000000000
			//   mov   rt, gt
			//   xorps rr, rt (or xorps rr, rt for abs computation)

			// NOTE: we can use allocTempReg, since we allocate from different classes,
			// AND all the called functions (nImmQ, nImmI) don't alloc/inspect the regstate
			// But this is arguably dangerous (some called function may change in the future),
			Register rt = allocTempReg(FpRegs & ~(rmask(ra)|rmask(rr)));
			Register gt = allocTempReg(GpRegs);

			if (ins->isop(LIR_absf) || ins->isop(LIR_absd)) {
				ANDPS(rr, rt);
			} else {
				XORPS(rr, rt);
			}

			if (ins->isop(LIR_negf)) {
				MOVDXR(rt, gt);
				nImmI(gt, negateMaskF[0], /*canClobberCCs*/true);
			} else if (ins->isop(LIR_negd)) {
				// LIR_negd
				MOVQXR(rt, gt);
				nImmQ(gt, negateMaskD[0], /*canClobberCCs*/true);
			} else if (ins->isop(LIR_absf)) {
				// LIR_absf
				MOVDXR(rt, gt);
				nImmI(gt, absMaskF[0], /*canClobberCCs*/true);
			} else if (ins->isop(LIR_absd)) {
				// LIR_negd
				MOVQXR(rt, gt);
				nImmQ(gt, absMaskD[0], /*canClobberCCs*/true);
			}
		}
		if (ra != rr)
			nNonGpCopy(rr,ra);
		endOpRegs(ins, rr, ra);
	}

	void Assembler::nRecipSqrt(LIns* ins) {

		// LIR_sqrt, LIR_rsqrt, LIR_recip

		Register rr, ra = UnspecifiedReg;   // init to shut GCC up
		beginOp1Regs(ins, FpRegs, rr, ra);

		switch (ins->opcode()) {
			case LIR_sqrtf:
				SQRTSS(rr, ra);
				break;
			case LIR_sqrtd:
				SQRTSD(rr, ra);
				break;
			default:
				NanoAssert(ins->opcode() == LIR_sqrtf || ins->opcode() == LIR_sqrtd); //Only sqrtf or sqrtd are managed at the moment
				break;
		}

		if (ra != rr)
			nNonGpCopy(rr, ra);
		endOpRegs(ins, rr, ra);
	}

	void Assembler::nSpill(const LIns* insToSpill) {

		int d = arDisp(insToSpill->getArIndex());
		NanoAssert(d >= 0);

		Register rr = insToSpill->getReg();
		LTy retType = insToSpill->retType();

		switch (retType)
		{
			case LTy_I:
				switch (insToSpill->getArSize()) {
					case 1:		MOVBMR(rr, d, FP); break;
					case 2:		MOVSMR(rr, d, FP); break;
					case 4:		MOVLMR(rr, d, FP); break;
					default:	NanoAssert(0); break;
				}
				break;

			case LTy_Q:
				MOVQMR(rr, d, FP);
				break;

			case LTy_F:
				// single-precision float: store 32bits from XMM to memory
				MOVSSMR(rr, d, FP);
				break;

			case LTy_D:
				// double: store 64bits from XMM to memory
				MOVSDMR(rr, d, FP);
				break;

			default:
				NanoAssert(0); // invalid retType
		}
	}

	NIns* Assembler::nEntryPoint(RegisterMask saveMask) {
		// reserve 64 bytes of stack space for external call parameters
		// add 8 to preserve 16 bytes alignment
		NanoAssert(saveMask == 0);
		NanoUnusedVariable(saveMask);

		SUBQR8(RSP, 0x48);
		return _nIns;
	}

	NIns* Assembler::nExitPoint(RegisterMask restoreMask) {

		NanoAssert(restoreMask == 0);
		NanoUnusedVariable(restoreMask);

		RET();
		ADDQR8(RSP, 0x48);
		return _nIns;
	}

	void Assembler::nPrologue() {

		// add
		for (FarJumpList::Iter it = _farJumpList->begin(), itEnd = _farJumpList->end(); it != itEnd; ++it) {

			const NIns*    target          = it->second;
			const size_t   sizeOfOffset    = sizeof(uint32_t);

			if (target == NULL)
				continue; // already processed

			underrunProtect(8);
			_nIns -= 8;
			*((const NIns**)_nIns) = target;

			for (FarJumpList::Iter itDuplicate = it; itDuplicate != itEnd; ++itDuplicate) {

				NIns*               duplicateOffsetPtr       = itDuplicate->first;
				const NIns*         duplicateTarget          = itDuplicate->second;

				if (duplicateTarget == target) {
					ptrdiff_t duplicateIpOffset = _nIns - (duplicateOffsetPtr + sizeOfOffset);
					NanoAssert(isS32(duplicateIpOffset));

					*((uint32_t*)duplicateOffsetPtr) = static_cast<uint32_t>(duplicateIpOffset);

					_relativeBranchRelocationList.push_back(makePair(duplicateOffsetPtr - oplen(X64_calli), static_cast<const NIns*>(_nIns)));

					itDuplicate->first = NULL;
					itDuplicate->second = NULL;
				}
			}
		}
	}


	void Assembler::nPatchRelativeBranch(NIns *patch, const NIns *target) {
		NIns *next = 0;

		if ((patch[0] == 0xFF) && (patch[1] == 0x15)) {
			// calli
			next = patch + 6;
		} else if (patch[0] == 0xE9) {
			// jmp disp32
			next = patch + 5;
		} else if ((patch[0] == 0x0F) && ((patch[1] & 0xF0) == 0x80)) {
			// jcc disp32
			next = patch + 6;
		} else {
			next = 0; // will cause a crash
			NanoAssertMsg(false, "cannot patch opcode");
		}

		// offset must fit in 32 bits
		if (!isS32(target - next)) {
			NanoAssertMsg(false, "distance between source and destination is too great");
			next = 0; // will cause a crash
		}

		((int32_t*)next)[-1] = int32_t(target - next);
	}

	void Assembler::nPatchRelativeOffset(const NIns *originalAddress, NIns *opcode) {
		NanoUnusedVariable(originalAddress);
		NanoUnusedVariable(opcode);
		NanoAssert(0); // not used
	}


	void Assembler::nCreate() {
	}

	void Assembler::nDestroy() {
	}

	void Assembler::nBeginAssembly(Allocator* compilationAllocator) {

		_farJumpList = new (compilationAllocator) FarJumpList(compilationAllocator);
		_farJumpList->reserve(64);
	}

	void Assembler::nEndAssembly(Allocator* compilationAllocator) {
		NanoUnusedVariable(compilationAllocator);
		_farJumpList = NULL;
	}

	void Assembler::nInsertRandomNop() {
		NanoAssert(0); // not supported
	}

	void Assembler::nLabel() {
		// do nothing right now
	}

	void Assembler::nCopyI(LIns* ins) {

		Register rr = UnspecifiedReg;
		Register ra = UnspecifiedReg;

		beginOp1Regs(ins, GpRegs, rr, ra);

		if (ins->getGenFlags() && ins->flagRequested()) {
			asm_force_gen_flags(ins);
		}

		if (rr != ra)
			emit_movi(rr, ra);
		endOpRegs(ins, rr, ra);

		freeResourcesOf(ins);
	}

	void Assembler::nCopyQ(LIns* ins) {

		Register rr = UnspecifiedReg;
		Register ra = UnspecifiedReg;

		beginOp1Regs(ins, GpRegs, rr, ra);

		if (ins->getGenFlags() && ins->flagRequested()) {
			asm_force_gen_flags(ins);
		}

		if (rr != ra)
			emit_movq(rr, ra);
		endOpRegs(ins, rr, ra);

		freeResourcesOf(ins);
	}

	void Assembler::nCopyF(LIns* ins) {

		Register rr = UnspecifiedReg;
		Register ra = UnspecifiedReg;

		if (ins->getGenFlags() && ins->flagRequested()) {
			asm_force_gen_flags(ins);
		}

		beginOp1Regs(ins, FpRegs, rr, ra);
		if (rr != ra)
			emit_movf(rr, ra);
		endOpRegs(ins, rr, ra);

		freeResourcesOf(ins);
	}

	void Assembler::nCopyD(LIns* ins) {

		Register rr = UnspecifiedReg;
		Register ra = UnspecifiedReg;

		if (ins->getGenFlags() && ins->flagRequested()) {
			asm_force_gen_flags(ins);
		}

		beginOp1Regs(ins, FpRegs, rr, ra);
		if (rr != ra)
			emit_movd(rr, ra);
		endOpRegs(ins, rr, ra);

		freeResourcesOf(ins);
	}

	void Assembler::emit_movi(Register r0, Register r1) {
		MOVLR(r0, r1);
	}

	void Assembler::emit_movq(Register r0, Register r1) {
		MOVQR(r0, r1);
	}

	void Assembler::emit_movf(Register r0, Register r1) {
		MOVAPSR(r0, r1);
	}

	void Assembler::emit_movd(Register r0, Register r1) {
		MOVAPSR(r0, r1);
	}

	void Assembler::nSaveFlag(LIns* ins) {
		LIns*       oprnd1  = ins->oprnd1();

		if (oprnd1->isImmAny() && oprnd1->canOptimize()) {
			asm_save_flag_imm(ins);
			return;
		}

		LOpcode     op  = ins->opcode();
		Register    rr  = prepareResultReg(ins, GpRegs);


		// force instantiation of operand
		findRegFor(oprnd1, GpRegs & ~rmask(rr));

		// zero extend to clear the upper bit of the register
		MOVZX8(rr, rr);
		switch (op) {
		case LIR_getflgz:       SETZ(rr);       break;
		case LIR_getflgs:       SETS(rr);       break;
		case LIR_getflgc:       SETC(rr);       break;
		case LIR_getflgb:       SETC(rr);       break;
		case LIR_getflgv:       SETO(rr);       break;
		case LIR_getflgnz:      SETNZ(rr);      break;
		case LIR_getflgns:      SETNS(rr);      break;
		case LIR_getflgnc:      SETNC(rr);      break;
		case LIR_getflgnb:      SETNC(rr);      break;
		case LIR_getflgnv:      SETNO(rr);      break;
		default:                NanoAssert(0);  break;// invalid opcode
		}

		freeResourcesOf(ins);
	}

	void Assembler::asm_save_flag_imm(LIns* ins) {
		LOpcode     op  = ins->opcode();
		Register    rr  = prepareResultReg(ins, GpRegs);
		LIns*       a   = ins->oprnd1();

		if (a->isImmI()) {
			uint32_t iImm = a->immI();

			switch (op) {
				case LIR_getflgz:       nImmI(rr, iImm == 0 ? 1 : 0, false);     break;
				case LIR_getflgs:       nImmI(rr, iImm >> 31, false);            break;
				case LIR_getflgc:       nImmI(rr, 0, false);                     break;
				case LIR_getflgv:       nImmI(rr, 0, false);                     break;
				case LIR_getflgnz:      nImmI(rr, iImm != 0 ? 1 : 0, false);     break;
				case LIR_getflgns:      nImmI(rr, 1 - (iImm >> 31), false);      break;
				case LIR_getflgnc:      nImmI(rr, 1, false);                     break;
				case LIR_getflgnv:      nImmI(rr, 1, false);                     break;
				default:                NanoAssert(0);                              break;
			}
		} else if (a->isImmQ()) {
			uint64_t iImm = a->immQ();

			switch (op) {
				case LIR_getflgz:       nImmQ(rr, iImm == 0 ? 1 : 0, false);     break;
				case LIR_getflgs:       nImmQ(rr, iImm >> 63, false);            break;
				case LIR_getflgc:       nImmQ(rr, 0, false);                     break;
				case LIR_getflgv:       nImmQ(rr, 0, false);                     break;
				case LIR_getflgnz:      nImmQ(rr, iImm != 0 ? 1 : 0, false);     break;
				case LIR_getflgns:      nImmQ(rr, 1 - (iImm >> 63), false);      break;
				case LIR_getflgnc:      nImmQ(rr, 1, false);                     break;
				case LIR_getflgnv:      nImmQ(rr, 1, false);                     break;
				default:                NanoAssert(0);                              break;
			}
		}

		freeResourcesOf(ins);
	}

	void Assembler::nRestoreFlag(LIns* ins) {
		LOpcode op = ins->opcode();
		Register ra = findRegFor(ins->oprnd1(), GpRegs);
		Register rr = allocTempReg(GpRegs & ~rmask(ra));

		switch (op) {
		case LIR_setflgz:       NanoAssert(0);  break;
		case LIR_setflgs:       NanoAssert(0);  break;
		case LIR_setflgc:       SHRLI(rr, 1);   break;
		case LIR_setflgb:       SHRLI(rr, 1);   break;
		case LIR_setflgv:       NanoAssert(0);  break;
		default:                NanoAssert(0);  break;// invalid opcode
		}

		MR(rr, ra);

		freeResourcesOf(ins);
	}

	void Assembler::asm_force_gen_flags(LIns* ins) {
		NanoAssert((ins->flagRequested() & (ArithFlag_Carry | ArithFlag_Borrow | ArithFlag_Overflow)) == 0);
		Register rr = ins->getReg();
		asm_force_gen_flags_reg(rr, ins->retType());
	}

	void Assembler::asm_force_gen_flags_reg(Register rr, LTy type) {
		if (type == LTy_I)
			ORLRR(rr, rr);
		else if (type == LTy_Q)
			ORQRR(rr, rr);
		else
			NanoAssert(0);

	}

	void Assembler::checkCondition(LIns* cond) {

#if defined(NANOJIT_ENABLE_ASSERT)
		switch (cond->opcode()) {
		case LIR_eqi:
		case LIR_nei:
		case LIR_lti:
		case LIR_gei:
		case LIR_gti:
		case LIR_lei:
		case LIR_ltui:
		case LIR_geui:
		case LIR_gtui:
		case LIR_leui:
			NanoAssert((cond->oprnd1()->retType() == LTy_I) && (cond->oprnd2()->retType() == LTy_I));
			break;
	
		case LIR_eqq:
		case LIR_neq:
		case LIR_ltq:
		case LIR_geq:
		case LIR_gtq:
		case LIR_leq:
		case LIR_ltuq:
		case LIR_geuq:
		case LIR_gtuq:
		case LIR_leuq:
			NanoAssert((cond->oprnd1()->retType() == LTy_Q) && (cond->oprnd2()->retType() == LTy_Q));
			break;
	
		case LIR_eqf:
		case LIR_nef:
		case LIR_ltf:
		case LIR_gef:
		case LIR_gtf:
		case LIR_lef:
		case LIR_unf:
			NanoAssert((cond->oprnd1()->retType() == LTy_F) && (cond->oprnd2()->retType() == LTy_F));
			break;
	
		case LIR_eqd:
		case LIR_ned:
		case LIR_ltd:
		case LIR_ged:
		case LIR_gtd:
		case LIR_led:
		case LIR_und:
			NanoAssert((cond->oprnd1()->retType() == LTy_D) && (cond->oprnd2()->retType() == LTy_D));
			break;

		default:
			NanoAssert(0);
			break;
		}
#else
		NanoUnusedVariable(cond);
#endif
	}

	void Assembler::emit_cmovf(LIns* ins) {
		RegisterMask allow = FpRegs;

		Register rr, rt, rf, rca, rcb, rtmp1;
		LIns* cond			= ins->oprnd1();
		LIns* iftrue		= ins->oprnd2();
		LIns* iffalse		= ins->oprnd3();
		LIns* condoprnd1	= cond->oprnd1();
		LIns* condoprnd2	= cond->oprnd2();
		LOpcode condop = cond->opcode();

		// condition operand must be set to live
		setResultLive(condoprnd1);
		setResultLive(condoprnd2);

		rr = prepareResultReg(ins, allow);
		allow &= ~rmask(rr);

		if (!iftrue->isInReg())
			rt = findRegFor(iftrue, FpRegs);
		else
			rt = iftrue->getReg();
		allow &= ~rmask(rt);

		if (!iffalse->isInReg())
			rf = findRegFor(iffalse, FpRegs);
		else
			rf = iffalse->getReg();
		allow &= ~rmask(rf);

		if (!condoprnd1->isInReg())
			rca = findRegFor(condoprnd1, FpRegs);
		else
			rca = condoprnd1->getReg();
		allow &= ~rmask(rca);

		if (!condoprnd2->isInReg())
			rcb = findRegFor(condoprnd2, FpRegs);
		else
			rcb = condoprnd2->getReg();
		allow &= ~rmask(rcb);

		rtmp1 = allocTempReg(allow);


		if (iftrue->retType() == LTy_D) {
			ORPD(rr, rtmp1);
			ANDPD(rtmp1, rt);
			ANDNPD(rr, rf);
			MOVAPSR(rtmp1, rr);
		} else if (iftrue->retType() == LTy_F) {
			ORPS(rr, rtmp1);
			ANDPS(rtmp1, rt);
			ANDNPS(rr, rf);
			MOVAPSR(rtmp1, rr);
		} else {
			NanoAssert(0);
		}

		switch (condop) {
			case LIR_eqf:			CMPEQSS   (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_nef:			CMPNEQSS  (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_ltf:			CMPLTSS   (rr, rcb);	MOVAPSR(rr, rca);	break;

			case LIR_gef:			CMPLESS   (rr, rca);	MOVAPSR(rr, rcb);	break;
			case LIR_gtf:			CMPLTSS   (rr, rca);	MOVAPSR(rr, rcb);	break;

			case LIR_lef:			CMPLESS   (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_unf:			CMPUNORDSS(rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_eqd:			CMPEQSD   (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_ned:			CMPNEQSD  (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_ltd:			CMPLTSD   (rr, rcb);	MOVAPSR(rr, rca);	break;

			case LIR_ged:			CMPLESD   (rr, rca);	MOVAPSR(rr, rcb);	break;
			case LIR_gtd:			CMPLTSD   (rr, rca);	MOVAPSR(rr, rcb);	break;

			case LIR_led:			CMPLESD   (rr, rcb);	MOVAPSR(rr, rca);	break;
			case LIR_und:			CMPUNORDSD(rr, rcb);	MOVAPSR(rr, rca);	break;

			default:				NanoAssert(0); /* invalid opcode */			break;
		}

		freeResourcesOf(ins);
	}



	// regalloc


	// Init per-opcode register hint table.
	void RegAlloc::nInitHint() {
		RegisterMask defaultFavorite = 0; //~(rmask(RAX) | rmask(RCX) | rmask(RDX) | rmask(RDI) | rmask(RSI));
		//VMPI_memset(Hints,0,sizeof(RegisterMask)*LIR_sentinel );
		for (unsigned int iOpcodeIndex = 0; iOpcodeIndex < LIR_sentinel; ++iOpcodeIndex) {
			_hint[iOpcodeIndex] = defaultFavorite;
		}

		_hint[LIR_calli]  = rmask(RAX);
		_hint[LIR_calld]  = rmask(XMM0);
		_hint[LIR_callf]  = rmask(XMM0);
		_hint[LIR_parami] = PREFER_SPECIAL;
#if defined(NANOJIT_64BIT)
		_hint[LIR_paramq] = PREFER_SPECIAL;
#endif
		_hint[LIR_paramf] = PREFER_SPECIAL;
		_hint[LIR_paramd] = PREFER_SPECIAL;
	}

	bool RegAlloc::nCanRemat(LIns* ins) {
		return (ins->canOptimize() && ins->isImmAny());
	}

	RegisterMask RegAlloc::nRegCopyCandidates(Register r, RegisterMask allow) {
		(void) r;
		return allow; // can freely transfer registers among different classes
	}

	RegisterMask RegAlloc::nInitManagedRegisters(const Config* config) {
		NanoUnusedVariable(config);
		return 0xffffffff & ~((1<<REGNUM(RBP)) | (1<<REGNUM(RSP)));
	}

	RegisterMask RegAlloc::nHint(LIns* ins) {
		NanoAssert(_hint[ins->opcode()] == PREFER_SPECIAL);
		NanoAssert(ins->isLInsP());

		RegisterMask prefer = 0;
		Register reg = ins->paramReg();
		if (reg != UnspecifiedReg)
			prefer = rmask(reg);

		return prefer;
	}


} // namespace nanojit

#endif // NANOJIT_HOST_PLATFORM_X64
